from _typeshed import Incomplete

unicode = str

class QR:
    valid: Incomplete
    bits: Incomplete
    group: int
    data: Incomplete
    def __init__(self, data) -> None: ...
    def __len__(self) -> int: ...
    @property
    def bitlength(self): ...
    def getLengthBits(self, ver): ...
    def getLength(self): ...
    def write_header(self, buffer, version) -> None: ...
    def write(self, buffer, version) -> None: ...

class QRNumber(QR):
    valid: Incomplete
    chars: str
    bits: Incomplete
    group: int
    mode: int
    lengthbits: Incomplete

class QRAlphaNum(QR):
    valid: Incomplete
    chars: str
    bits: Incomplete
    group: int
    mode: int
    lengthbits: Incomplete

class QR8bitByte(QR):
    bits: Incomplete
    group: int
    mode: int
    lengthbits: Incomplete
    data: Incomplete
    def __init__(self, data) -> None: ...
    def write(self, buffer, version) -> None: ...

class QRKanji(QR):
    bits: Incomplete
    group: int
    mode: int
    lengthbits: Incomplete
    data: Incomplete
    def __init__(self, data) -> None: ...
    def unicode_to_qrkanji(self, data): ...
    def write(self, buffer, version) -> None: ...

class QRHanzi(QR):
    bits: Incomplete
    group: int
    mode: int
    lengthbits: Incomplete
    data: Incomplete
    def __init__(self, data) -> None: ...
    def unicode_to_qrhanzi(self, data): ...
    def write_header(self, buffer, version) -> None: ...
    def write(self, buffer, version) -> None: ...

class QRECI(QR):
    mode: int
    lengthbits: Incomplete
    data: Incomplete
    def __init__(self, data) -> None: ...
    def write(self, buffer, version) -> None: ...

class QRStructAppend(QR):
    mode: int
    lengthbits: Incomplete
    part: Incomplete
    total: Incomplete
    parity: Incomplete
    def __init__(self, part, total, parity) -> None: ...
    def write(self, buffer, version) -> None: ...

class QRFNC1First(QR):
    mode: int
    lengthbits: Incomplete
    def __init__(self) -> None: ...
    def write(self, buffer, version) -> None: ...

class QRFNC1Second(QR):
    valid: Incomplete
    mode: int
    lengthbits: Incomplete
    def write(self, buffer, version) -> None: ...

class QRCode:
    version: Incomplete
    errorCorrectLevel: Incomplete
    modules: Incomplete
    moduleCount: int
    dataCache: Incomplete
    dataList: Incomplete
    def __init__(self, version, errorCorrectLevel) -> None: ...
    def addData(self, data) -> None: ...
    def isDark(self, row, col): ...
    def getModuleCount(self): ...
    def calculate_version(self): ...
    def make(self) -> None: ...
    def makeImpl(self, test, maskPattern) -> None: ...
    def setupPositionProbePattern(self, row, col) -> None: ...
    def getBestMaskPattern(self): ...
    def setupTimingPattern(self) -> None: ...
    def setupPositionAdjustPattern(self) -> None: ...
    def setupTypeNumber(self, test) -> None: ...
    def setupTypeInfo(self, test, maskPattern) -> None: ...
    def dataPosIterator(self): ...
    def dataBitIterator(self, data): ...
    def mapData(self, data, maskPattern) -> None: ...
    PAD0: int
    PAD1: int
    @staticmethod
    def createData(version, errorCorrectLevel, dataList): ...
    @staticmethod
    def createBytes(buffer, rsBlocks): ...

class QRErrorCorrectLevel:
    L: int
    M: int
    Q: int
    H: int

class QRMaskPattern:
    PATTERN000: int
    PATTERN001: int
    PATTERN010: int
    PATTERN011: int
    PATTERN100: int
    PATTERN101: int
    PATTERN110: int
    PATTERN111: int

class QRUtil:
    PATTERN_POSITION_TABLE: Incomplete
    G15: Incomplete
    G18: Incomplete
    G15_MASK: Incomplete
    @staticmethod
    def getBCHTypeInfo(data): ...
    @staticmethod
    def getBCHTypeNumber(data): ...
    @staticmethod
    def getBCHDigit(data): ...
    @staticmethod
    def getPatternPosition(version): ...
    maskPattern: Incomplete
    @classmethod
    def getMask(cls, maskPattern): ...
    @staticmethod
    def getErrorCorrectPolynomial(errorCorrectLength): ...
    @classmethod
    def maskScoreRule1vert(cls, modules): ...
    @classmethod
    def maskScoreRule2(cls, modules): ...
    @classmethod
    def maskScoreRule3hor(cls, modules, pattern=...): ...
    @classmethod
    def maskScoreRule4(cls, modules): ...
    @classmethod
    def getLostPoint(cls, qrCode): ...

class QRMath:
    @staticmethod
    def glog(n): ...
    @staticmethod
    def gexp(n): ...

EXP_TABLE: Incomplete
LOG_TABLE: Incomplete

class QRPolynomial:
    num: Incomplete
    def __init__(self, num, shift) -> None: ...
    def get(self, index): ...
    def getLength(self): ...
    def multiply(self, e): ...
    def mod(self, e): ...

class QRRSBlock:
    RS_BLOCK_TABLE: Incomplete
    totalCount: Incomplete
    dataCount: Incomplete
    def __init__(self, totalCount, dataCount) -> None: ...
    @staticmethod
    def getRSBlocks(version, errorCorrectLevel): ...
    @staticmethod
    def getRsBlockTable(version, errorCorrectLevel): ...

class QRBitBuffer:
    buffer: Incomplete
    length: int
    def __init__(self) -> None: ...
    def get(self, index): ...
    def put(self, num, length) -> None: ...
    def getLengthInBits(self): ...
    def putBit(self, bit) -> None: ...



import enum
import sys
from collections.abc import Callable
from types import TracebackType
from typing import Any
from typing_extensions import Self

from pynput._util import AbstractListener

class Button(enum.Enum):
    unknown = 0
    left = 1
    middle = 2
    right = 3
    if sys.platform == "linux":
        button8 = 8
        button9 = 9
        button10 = 10
        button11 = 11
        button12 = 12
        button13 = 13
        button14 = 14
        button15 = 15
        button16 = 16
        button17 = 17
        button18 = 18
        button19 = 19
        button20 = 20
        button21 = 21
        button22 = 22
        button23 = 23
        button24 = 24
        button25 = 25
        button26 = 26
        button27 = 27
        button28 = 28
        button29 = 29
        button30 = 30
        scroll_down = 5
        scroll_left = 6
        scroll_right = 7
        scroll_up = 4
    if sys.platform == "win32":
        x1 = 0  # Value unknown
        x2 = 0  # Value unknown

class Controller:
    def __init__(self) -> None: ...
    @property
    def position(self) -> tuple[int, int]: ...
    @position.setter
    def position(self, position: tuple[int, int]) -> None: ...
    def scroll(self, dx: int, dy: int) -> None: ...
    def press(self, button: Button) -> None: ...
    def release(self, button: Button) -> None: ...
    def move(self, dx: int, dy: int) -> None: ...
    def click(self, button: Button, count: int = 1) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> None: ...

class Listener(AbstractListener):
    if sys.platform == "win32":
        WM_LBUTTONDOWN: int
        WM_LBUTTONUP: int
        WM_MBUTTONDOWN: int
        WM_MBUTTONUP: int
        WM_MOUSEMOVE: int
        WM_MOUSEWHEEL: int
        WM_MOUSEHWHEEL: int
        WM_RBUTTONDOWN: int
        WM_RBUTTONUP: int
        WM_XBUTTONDOWN: int
        WM_XBUTTONUP: int

        MK_XBUTTON1: int
        MK_XBUTTON2: int

        XBUTTON1: int
        XBUTTON2: int

        CLICK_BUTTONS: dict[int, tuple[Button, bool]]
        X_BUTTONS: dict[int, dict[int, tuple[Button, bool]]]
        SCROLL_BUTTONS: dict[int, tuple[int, int]]

    def __init__(
        self,
        on_move: Callable[[int, int], bool | None] | None = None,
        on_click: Callable[[int, int, Button, bool], bool | None] | None = None,
        on_scroll: Callable[[int, int, int, int], bool | None] | None = None,
        suppress: bool = False,
        **kwargs: Any,
    ) -> None: ...



from _typeshed import Incomplete

from authlib.integrations.base_client import FrameworkIntegration

__all__ = ["BaseOAuth"]

class BaseOAuth:
    oauth1_client_cls: Incomplete
    oauth2_client_cls: Incomplete
    framework_integration_cls: type[FrameworkIntegration] = ...
    cache: Incomplete
    fetch_token: Incomplete
    update_token: Incomplete
    def __init__(self, cache=None, fetch_token=None, update_token=None) -> None: ...
    def create_client(self, name): ...
    def register(self, name, overwrite: bool = False, **kwargs): ...
    def generate_client_kwargs(self, name, overwrite, **kwargs): ...
    def load_config(self, name, params): ...
    def __getattr__(self, key): ...



from _typeshed import Incomplete
from typing import Final

__version__: Final[str]

class Logger:
    def __init__(self) -> None: ...
    def add(self, fp) -> None: ...
    def remove(self, fp) -> None: ...
    def write(self, text) -> None: ...
    def __call__(self, text) -> None: ...

logger: Incomplete

class WarnOnce:
    uttered: Incomplete
    pfx: Incomplete
    enabled: int
    def __init__(self, kind: str = "Warn") -> None: ...
    def once(self, warning) -> None: ...
    def __call__(self, warning) -> None: ...

warnOnce: Incomplete
infoOnce: Incomplete



from reportlab.lib.styles import PropertySet
from reportlab.platypus.paragraph import Paragraph, ParaLines
from reportlab.platypus.paraparser import ParaFrag

class XPreformatted(Paragraph):
    def __init__(
        self,
        text: str,
        # NOTE: This should be a ParagraphStyle
        style: PropertySet,
        bulletText: str | None = None,
        frags: list[ParaFrag] | None = None,
        caseSensitive: int = 1,
        dedent: int = 0,
    ) -> None: ...
    def breakLinesCJK(self, width: float | list[float] | tuple[float, ...]) -> ParaLines | ParaFrag: ...

class PythonPreformatted(XPreformatted):
    formats: dict[str, tuple[str, str]]
    def __init__(
        self,
        text: str,
        # NOTE: This should be a ParagraphStyle
        style: PropertySet,
        bulletText: str | None = None,
        dedent: int = 0,
        frags: list[ParaFrag] | None = None,
    ) -> None: ...
    def escapeHtml(self, text: str) -> str: ...
    def fontify(self, code: str) -> str: ...

__all__ = ("XPreformatted", "PythonPreformatted")



import io
import socket
from _typeshed import ReadableBuffer
from collections.abc import Iterable, Iterator

class Unreader:
    buf: io.BytesIO

    def __init__(self) -> None: ...
    def chunk(self) -> bytes: ...
    def read(self, size: int | None = None) -> bytes: ...
    def unread(self, data: ReadableBuffer) -> None: ...

class SocketUnreader(Unreader):
    sock: socket.socket
    mxchunk: int

    def __init__(self, sock: socket.socket, max_chunk: int = 8192) -> None: ...
    def chunk(self) -> bytes: ...

class IterUnreader(Unreader):
    iter: Iterator[bytes] | None

    def __init__(self, iterable: Iterable[bytes]) -> None: ...
    def chunk(self) -> bytes: ...



from ._base import Trie as ABCTrie

class Trie(ABCTrie):
    def __init__(self, data) -> None: ...
    def __contains__(self, key): ...
    def __len__(self) -> int: ...
    def __iter__(self): ...
    def __getitem__(self, key): ...
    def keys(self, prefix=None): ...
    def has_keys_with_prefix(self, prefix): ...



from typing import Any

class ExtendedOperationContainer:
    def __init__(self, connection) -> None: ...

class StandardExtendedOperations(ExtendedOperationContainer):
    def who_am_i(self, controls=None): ...
    def modify_password(self, user=None, old_password=None, new_password=None, hash_algorithm=None, salt=None, controls=None): ...
    def paged_search(
        self,
        search_base,
        search_filter,
        search_scope="SUBTREE",
        dereference_aliases="ALWAYS",
        attributes=None,
        size_limit: int = 0,
        time_limit: int = 0,
        types_only: bool = False,
        get_operational_attributes: bool = False,
        controls=None,
        paged_size: int = 100,
        paged_criticality: bool = False,
        generator: bool = True,
    ): ...
    def persistent_search(
        self,
        search_base: str = "",
        search_filter: str = "(objectclass=*)",
        search_scope="SUBTREE",
        dereference_aliases="NEVER",
        attributes="*",
        size_limit: int = 0,
        time_limit: int = 0,
        controls=None,
        changes_only: bool = True,
        show_additions: bool = True,
        show_deletions: bool = True,
        show_modifications: bool = True,
        show_dn_modifications: bool = True,
        notifications: bool = True,
        streaming: bool = True,
        callback=None,
    ): ...
    def funnel_search(
        self,
        search_base: str = "",
        search_filter: str = "",
        search_scope="SUBTREE",
        dereference_aliases="NEVER",
        attributes="*",
        size_limit: int = 0,
        time_limit: int = 0,
        controls=None,
        streaming: bool = False,
        callback=None,
    ): ...

class NovellExtendedOperations(ExtendedOperationContainer):
    def get_bind_dn(self, controls=None): ...
    def get_universal_password(self, user, controls=None): ...
    def set_universal_password(self, user, new_password=None, controls=None): ...
    def list_replicas(self, server_dn, controls=None): ...
    def partition_entry_count(self, partition_dn, controls=None): ...
    def replica_info(self, server_dn, partition_dn, controls=None): ...
    def start_transaction(self, controls=None): ...
    def end_transaction(self, commit: bool = True, controls=None): ...
    def add_members_to_groups(self, members, groups, fix: bool = True, transaction: bool = True): ...
    def remove_members_from_groups(self, members, groups, fix: bool = True, transaction: bool = True): ...
    def check_groups_memberships(self, members, groups, fix: bool = False, transaction: bool = True): ...

class MicrosoftExtendedOperations(ExtendedOperationContainer):
    def dir_sync(
        self,
        sync_base,
        sync_filter: str = "(objectclass=*)",
        attributes="*",
        cookie=None,
        object_security: bool = False,
        ancestors_first: bool = True,
        public_data_only: bool = False,
        incremental_values: bool = True,
        max_length: int = 2147483647,
        hex_guid: bool = False,
    ): ...
    def modify_password(self, user, new_password, old_password=None, controls=None): ...
    def unlock_account(self, user): ...
    def add_members_to_groups(self, members, groups, fix: bool = True): ...
    def remove_members_from_groups(self, members, groups, fix: bool = True): ...
    def persistent_search(
        self, search_base: str = "", search_scope="SUBTREE", attributes="*", streaming: bool = True, callback=None
    ): ...

class ExtendedOperationsRoot(ExtendedOperationContainer):
    standard: Any
    novell: Any
    microsoft: Any
    def __init__(self, connection) -> None: ...



from __future__ import annotations

from wtforms import Field, Form


class Filter1:
    def __call__(self, value: object) -> None: ...


class Filter2:
    def __call__(self, input: None) -> None: ...


def not_a_filter(a: object, b: object) -> None: ...


def also_not_a_filter() -> None: ...


# we should accept any mapping of sequences, we can't really validate
# the filter functions when it's this nested
form = Form()
form.process(extra_filters={"foo": (str.upper, str.strip, int), "bar": (Filter1(), Filter2())})
form.process(extra_filters={"foo": [str.upper, str.strip, int], "bar": [Filter1(), Filter2()]})

# regardless of how we pass the filters into Field it should work
field = Field(filters=(str.upper, str.lower, int))
Field(filters=(Filter1(), Filter2()))
Field(filters=[str.upper, str.lower, int])
Field(filters=[Filter1(), Filter2()])
field.process(None, extra_filters=(str.upper, str.lower, int))
field.process(None, extra_filters=(Filter1(), Filter2()))
field.process(None, extra_filters=[str.upper, str.lower, int])
field.process(None, extra_filters=[Filter1(), Filter2()])

# but if we pass in some callables with an incompatible param spec
# then we should get type errors
Field(filters=(str.upper, str.lower, int, not_a_filter))  # type: ignore
Field(filters=(Filter1(), Filter2(), also_not_a_filter))  # type: ignore
Field(filters=[str.upper, str.lower, int, also_not_a_filter])  # type: ignore
Field(filters=[Filter1(), Filter2(), not_a_filter])  # type: ignore
field.process(None, extra_filters=(str.upper, str.lower, int, not_a_filter))  # type: ignore
field.process(None, extra_filters=(Filter1(), Filter2(), also_not_a_filter))  # type: ignore
field.process(None, extra_filters=[str.upper, str.lower, int, also_not_a_filter])  # type: ignore
field.process(None, extra_filters=[Filter1(), Filter2(), not_a_filter])  # type: ignore



from typing import Protocol, TypeVar, overload
from typing_extensions import TypeAlias

_T = TypeVar("_T")
_GetterReturnType_co = TypeVar("_GetterReturnType_co", covariant=True)
_SetterValueType_contra = TypeVar("_SetterValueType_contra", contravariant=True)

class AsymmetricProperty(Protocol[_GetterReturnType_co, _SetterValueType_contra]):
    @overload
    def __get__(self, obj: None, type: type[object] | None = ..., /) -> property: ...
    @overload
    def __get__(self, obj: object, type: type[object] | None = ..., /) -> _GetterReturnType_co: ...
    def __set__(self, obj: object, value: _SetterValueType_contra, /) -> None: ...

class AsymmetricPropertyWithDelete(
    AsymmetricProperty[_GetterReturnType_co, _SetterValueType_contra], Protocol[_GetterReturnType_co, _SetterValueType_contra]
):
    def __delete__(self, obj: object, /) -> None: ...

SymmetricProperty: TypeAlias = AsymmetricProperty[_T, _T]
SymmetricPropertyWithDelete: TypeAlias = AsymmetricPropertyWithDelete[_T, _T]



import logging
from collections.abc import Sequence
from re import Pattern
from typing import Final, NamedTuple

COMMA_SEPARATED_LIST_RE: Final[Pattern[str]]
LOCAL_PLUGIN_LIST_RE: Final[Pattern[str]]
NORMALIZE_PACKAGE_NAME_RE: Final[Pattern[str]]

def parse_comma_separated_list(value: str, regexp: Pattern[str] = ...) -> list[str]: ...

class _Token(NamedTuple):
    tp: str
    src: str

def parse_files_to_codes_mapping(value_: Sequence[str] | str) -> list[tuple[str, list[str]]]: ...
def normalize_paths(paths: Sequence[str], parent: str = ".") -> list[str]: ...
def normalize_path(path: str, parent: str = ".") -> str: ...
def stdin_get_value() -> str: ...
def stdin_get_lines() -> list[str]: ...
def is_using_stdin(paths: list[str]) -> bool: ...
def fnmatch(filename: str, patterns: Sequence[str]) -> bool: ...
def matches_filename(path: str, patterns: Sequence[str], log_message: str, logger: logging.Logger) -> bool: ...
def get_python_version() -> str: ...
def normalize_pypi_name(s: str) -> str: ...



import sys
from _typeshed import SupportsWrite
from typing import Final

from .timemachine import *

SIGNATURE: Final[bytes]
EOCSID: Final[int]
FREESID: Final[int]
SATSID: Final[int]
MSATSID: Final[int]
EVILSID: Final[int]

class CompDocError(Exception): ...

class DirNode:
    DID: int
    name: str
    etype: int
    colour: int
    left_DID: int
    right_DID: int
    root_DID: int
    first_SID: int
    tot_size: int
    children: list[int]
    parent: int
    tsinfo: tuple[int, int, int, int]
    logfile: SupportsWrite[str]
    def __init__(self, DID: int, dent: bytes, DEBUG: int = 0, logfile: SupportsWrite[str] = sys.stdout) -> None: ...
    def dump(self, DEBUG: int = 1) -> None: ...

class CompDoc:
    logfile: SupportsWrite[str]
    ignore_workbook_corruption: bool
    DEBUG: int
    mem: bytes
    sec_size: int
    short_sec_size: int
    mem_data_secs: int
    mem_data_len: int
    seen: list[int]
    SAT: list[int]
    dirlist: list[DirNode]
    SSCS: str
    SSAT: list[int]
    def __init__(
        self, mem: bytes, logfile: SupportsWrite[str] = sys.stdout, DEBUG: int = 0, ignore_workbook_corruption: bool = False
    ) -> None: ...
    def get_named_stream(self, qname: str) -> bytes | None: ...
    def locate_named_stream(self, qname: str) -> tuple[bytes | None, int, int]: ...

def x_dump_line(alist: list[int], stride: int, f: SupportsWrite[str], dpos: int, equal: int = 0) -> None: ...
def dump_list(alist: list[int], stride: int, f: SupportsWrite[str] = sys.stdout) -> None: ...



from _typeshed import Incomplete
from typing import Final

__version__: Final[str]

def mkTimeTuple(timeString): ...
def str2seconds(timeString): ...
def seconds2str(seconds): ...
def nextRoundNumber(x): ...
def find_interval(lo, hi, I: int = 5): ...
def find_good_grid(lower, upper, n=(4, 5, 6, 7, 8, 9), grid=None): ...
def ticks(lower, upper, n=(4, 5, 6, 7, 8, 9), split: int = 1, percent: int = 0, grid=None, labelVOffset: int = 0): ...
def findNones(data): ...
def pairFixNones(pairs): ...
def maverage(data, n: int = 6): ...
def pairMaverage(data, n: int = 6): ...

class DrawTimeCollector:
    formats: Incomplete
    disabled: bool
    def __init__(self, formats=["gif"]) -> None: ...
    def clear(self) -> None: ...
    def record(self, func, node, *args, **kwds) -> None: ...
    def __call__(self, node, canvas, renderer) -> None: ...
    @staticmethod
    def rectDrawTimeCallback(node, canvas, renderer, **kwds): ...
    @staticmethod
    def transformAndFlatten(A, p): ...
    @property
    def pmcanv(self): ...
    def wedgeDrawTimeCallback(self, node, canvas, renderer, **kwds): ...
    def save(self, fnroot) -> None: ...

def xyDist(xxx_todo_changeme, xxx_todo_changeme1): ...
def lineSegmentIntersect(xxx_todo_changeme2, xxx_todo_changeme3, xxx_todo_changeme4, xxx_todo_changeme5): ...
def makeCircularString(x, y, radius, angle, text, fontName, fontSize, inside: int = 0, G=None, textAnchor: str = "start"): ...

class CustomDrawChanger:
    store: Incomplete
    def __init__(self) -> None: ...
    def __call__(self, change, obj) -> None: ...

class FillPairedData(list[Incomplete]):
    other: Incomplete
    def __init__(self, v, other: int = 0) -> None: ...



from _typeshed import Incomplete
from typing import ClassVar, Literal

import _win32typing

def odbc(connectionString: str, /) -> _win32typing.connection: ...
def SQLDataSources(direction, /) -> tuple[Incomplete, Incomplete]: ...

DATE: str
NUMBER: str
RAW: str
SQL_FETCH_ABSOLUTE: int
SQL_FETCH_FIRST: int
SQL_FETCH_FIRST_SYSTEM: int
SQL_FETCH_FIRST_USER: int
SQL_FETCH_LAST: int
SQL_FETCH_NEXT: int
SQL_FETCH_PRIOR: int
SQL_FETCH_RELATIVE: int
STRING: str
TYPES: tuple[Literal["STRING"], Literal["RAW"], Literal["NUMBER"], Literal["DATE"]]

class error(Exception):
    __name__: ClassVar[str] = "odbcError"

# These all pretend to come from a module called "dbi", but that module doesn't exist
class dataError(Exception): ...
class integrityError(Exception): ...
class internalError(Exception): ...
class noError(Exception): ...
class opError(Exception): ...
class progError(Exception): ...



#!/usr/bin/env python3
"""Run mypy on the test cases for the stdlib and third-party stubs."""

from __future__ import annotations

import argparse
import concurrent.futures
import os
import queue
import re
import shutil
import subprocess
import sys
import tempfile
import threading
from collections.abc import Callable, Generator
from contextlib import ExitStack, contextmanager, suppress
from dataclasses import dataclass
from enum import IntEnum
from functools import partial
from pathlib import Path
from typing_extensions import TypeAlias

from ts_utils.metadata import get_recursive_requirements, read_metadata
from ts_utils.mypy import mypy_configuration_from_distribution, temporary_mypy_config_file
from ts_utils.paths import STDLIB_PATH, TEST_CASES_DIR, TS_BASE_PATH, distribution_path
from ts_utils.utils import (
    PYTHON_VERSION,
    DistributionTests,
    colored,
    distribution_info,
    get_all_testcase_directories,
    get_mypy_req,
    print_error,
    venv_python,
)

ReturnCode: TypeAlias = int

VENV_DIR = ".venv"
TYPESHED = "typeshed"

SUPPORTED_PLATFORMS = ["linux", "darwin", "win32"]
SUPPORTED_VERSIONS = ["3.14", "3.13", "3.12", "3.11", "3.10", "3.9"]


def distribution_with_test_cases(distribution_name: str) -> DistributionTests:
    """Parse a CLI argument that is intended to be to a valid typeshed distribution."""
    try:
        return distribution_info(distribution_name)
    except RuntimeError as exc:
        raise argparse.ArgumentTypeError(str(exc)) from exc


class Verbosity(IntEnum):
    QUIET = 0
    NORMAL = 1
    VERBOSE = 2


parser = argparse.ArgumentParser(description="Script to run mypy against various test cases for typeshed's stubs")
parser.add_argument(
    "packages_to_test",
    type=distribution_with_test_cases,
    nargs="*",
    action="extend",
    help=(
        "Test only these packages (defaults to all typeshed stubs that have test cases). "
        'Use "stdlib" to test the standard library test cases.'
    ),
)
parser.add_argument(
    "--all",
    action="store_true",
    help=(
        'Run tests on all available platforms and versions (defaults to "False"). '
        "Note that this cannot be specified if --platform and/or --python-version are specified."
    ),
)
parser.add_argument(
    "--verbosity",
    choices=[member.name for member in Verbosity],
    default=Verbosity.NORMAL.name,
    help="Control how much output to print to the terminal",
)
parser.add_argument(
    "--platform",
    dest="platforms_to_test",
    choices=SUPPORTED_PLATFORMS,
    nargs="*",
    action="extend",
    help=(
        "Run mypy for certain OS platforms (defaults to sys.platform). "
        "Note that this cannot be specified if --all is also specified."
    ),
)
parser.add_argument(
    "-p",
    "--python-version",
    dest="versions_to_test",
    choices=SUPPORTED_VERSIONS,
    nargs="*",
    action="extend",
    help=(
        "Run mypy for certain Python versions (defaults to sys.version_info[:2]). "
        "Note that this cannot be specified if --all is also specified."
    ),
)

_PRINT_QUEUE: queue.SimpleQueue[str] = queue.SimpleQueue()


def verbose_log(msg: str) -> None:
    _PRINT_QUEUE.put(colored(msg, "blue"))


def setup_testcase_dir(package: DistributionTests, tempdir: Path, verbosity: Verbosity) -> None:
    if verbosity is verbosity.VERBOSE:
        verbose_log(f"{package.name}: Setting up testcase dir in {tempdir}")
    # --warn-unused-ignores doesn't work for files inside typeshed.
    # SO, to work around this, we copy the test_cases directory into a TemporaryDirectory,
    # and run the test cases inside of that.
    shutil.copytree(package.test_cases_path, tempdir / TEST_CASES_DIR)
    if package.is_stdlib:
        return

    # HACK: we want to run these test cases in an isolated environment --
    # we want mypy to see all stub packages listed in the "requires" field of METADATA.toml
    # (and all stub packages required by those stub packages, etc. etc.),
    # but none of the other stubs in typeshed.
    #
    # The best way of doing that without stopping --warn-unused-ignore from working
    # seems to be to create a "new typeshed" directory in a tempdir
    # that has only the required stubs copied over.
    new_typeshed = tempdir / TYPESHED
    new_typeshed.mkdir()
    shutil.copytree(STDLIB_PATH, new_typeshed / "stdlib")
    requirements = get_recursive_requirements(package.name)
    # mypy refuses to consider a directory a "valid typeshed directory"
    # unless there's a stubs/mypy-extensions path inside it,
    # so add that to the list of stubs to copy over to the new directory
    typeshed_requirements = [r.name for r in requirements.typeshed_pkgs]
    for requirement in {package.name, *typeshed_requirements, "mypy-extensions"}:
        shutil.copytree(distribution_path(requirement), new_typeshed / "stubs" / requirement)

    if requirements.external_pkgs:
        venv_location = str(tempdir / VENV_DIR)
        subprocess.run(["uv", "venv", venv_location], check=True, capture_output=True)
        ext_requirements = [str(r) for r in requirements.external_pkgs]
        uv_command = ["uv", "pip", "install", get_mypy_req(), *ext_requirements]
        if sys.platform == "win32":
            # Reads/writes to the cache are threadsafe with uv generally...
            # but not on old Windows versions
            # https://github.com/astral-sh/uv/issues/2810
            uv_command.append("--no-cache-dir")
        if verbosity is Verbosity.VERBOSE:
            verbose_log(f"{package.name}: Setting up venv in {venv_location}. {uv_command=}\n")
        try:
            subprocess.run(
                uv_command, check=True, capture_output=True, text=True, env=os.environ | {"VIRTUAL_ENV": venv_location}
            )
        except subprocess.CalledProcessError as e:
            _PRINT_QUEUE.put(f"{package.name}\n{e.stderr}")
            raise


def run_testcases(
    package: DistributionTests, version: str, platform: str, *, tempdir: Path, verbosity: Verbosity
) -> subprocess.CompletedProcess[str]:
    env_vars = dict(os.environ)
    new_test_case_dir = tempdir / TEST_CASES_DIR

    if package.is_stdlib:
        configurations = []
    else:
        configurations = mypy_configuration_from_distribution(package.name)

    with temporary_mypy_config_file(configurations) as temp:

        # "--enable-error-code ignore-without-code" is purposefully omitted.
        # See https://github.com/python/typeshed/pull/8083
        flags = [
            "--python-version",
            version,
            "--show-traceback",
            "--no-error-summary",
            "--platform",
            platform,
            "--strict",
            "--pretty",
            "--config-file",
            temp.name,
            # Avoid race conditions when reading the cache
            # (https://github.com/python/typeshed/issues/11220)
            "--no-incremental",
            # Not useful for the test cases
            "--disable-error-code=empty-body",
        ]

        if package.is_stdlib:
            python_exe = sys.executable
            custom_typeshed = TS_BASE_PATH
            flags.append("--no-site-packages")
        else:
            custom_typeshed = tempdir / TYPESHED
            env_vars["MYPYPATH"] = os.pathsep.join(map(str, custom_typeshed.glob("stubs/*")))
            has_non_types_dependencies = (tempdir / VENV_DIR).exists()
            if has_non_types_dependencies:
                python_exe = str(venv_python(tempdir / VENV_DIR))
            else:
                python_exe = sys.executable
                flags.append("--no-site-packages")

        flags.extend(["--custom-typeshed-dir", str(custom_typeshed)])

        # If the test-case filename ends with -py39,
        # only run the test if --python-version was set to 3.9 or higher (for example)
        for path in new_test_case_dir.rglob("*.py"):
            if match := re.fullmatch(r".*-py3(\d{1,2})", path.stem):
                minor_version_required = int(match[1])
                assert f"3.{minor_version_required}" in SUPPORTED_VERSIONS
                python_minor_version = int(version.split(".")[1])
                if minor_version_required > python_minor_version:
                    continue
            flags.append(str(path))

        mypy_command = [python_exe, "-m", "mypy", *flags]
        if verbosity is Verbosity.VERBOSE:
            description = f"{package.name}/{version}/{platform}"
            msg = f"{description}: {mypy_command=}\n"
            if "MYPYPATH" in env_vars:
                msg += f"{description}: {env_vars['MYPYPATH']=}"
            else:
                msg += f"{description}: MYPYPATH not set"
            msg += "\n"
            verbose_log(msg)
        return subprocess.run(mypy_command, capture_output=True, text=True, env=env_vars, check=False)


@dataclass(frozen=True)
class Result:
    code: int
    command_run: str
    stderr: str
    stdout: str
    test_case_dir: Path
    tempdir: Path

    def print_description(self) -> None:
        if self.code:
            print(f"{self.command_run}:", end=" ")
            print_error("FAILURE\n")
            replacements = (str(self.tempdir / TEST_CASES_DIR), str(self.test_case_dir))
            if self.stderr:
                print_error(self.stderr, fix_path=replacements)
            if self.stdout:
                print_error(self.stdout, fix_path=replacements)


def test_testcase_directory(
    package: DistributionTests, version: str, platform: str, *, verbosity: Verbosity, tempdir: Path
) -> Result:
    msg = f"mypy --platform {platform} --python-version {version} on the "
    msg += "standard library test cases" if package.is_stdlib else f"test cases for {package.name!r}"
    if verbosity > Verbosity.QUIET:
        _PRINT_QUEUE.put(f"Running {msg}...")

    proc_info = run_testcases(package=package, version=version, platform=platform, tempdir=tempdir, verbosity=verbosity)
    return Result(
        code=proc_info.returncode,
        command_run=msg,
        stderr=proc_info.stderr,
        stdout=proc_info.stdout,
        test_case_dir=package.test_cases_path,
        tempdir=tempdir,
    )


def print_queued_messages(ev: threading.Event) -> None:
    while not ev.is_set():
        with suppress(queue.Empty):
            print(_PRINT_QUEUE.get(timeout=0.5), flush=True)
    while True:
        try:
            msg = _PRINT_QUEUE.get_nowait()
        except queue.Empty:
            return
        else:
            print(msg, flush=True)


def concurrently_run_testcases(
    stack: ExitStack,
    testcase_directories: list[DistributionTests],
    verbosity: Verbosity,
    platforms_to_test: list[str],
    versions_to_test: list[str],
) -> list[Result]:
    packageinfo_to_tempdir = {
        distribution_info: Path(stack.enter_context(tempfile.TemporaryDirectory())) for distribution_info in testcase_directories
    }
    to_do: list[Callable[[], Result]] = []
    for testcase_dir, tempdir in packageinfo_to_tempdir.items():
        pkg = testcase_dir.name
        requires_python = None
        if not testcase_dir.is_stdlib:
            requires_python = read_metadata(pkg).requires_python
            if not requires_python.contains(PYTHON_VERSION):
                msg = f"skipping {pkg!r} (requires Python {requires_python}; test is being run using Python {PYTHON_VERSION})"
                print(colored(msg, "yellow"))
                continue
        for version in versions_to_test:
            if not testcase_dir.is_stdlib:
                assert requires_python is not None
                if not requires_python.contains(version):
                    msg = f"skipping {pkg!r} for target Python {version} (requires Python {requires_python})"
                    print(colored(msg, "yellow"))
                    continue
            to_do.extend(
                partial(test_testcase_directory, testcase_dir, version, platform, verbosity=verbosity, tempdir=tempdir)
                for platform in platforms_to_test
            )

    if not to_do:
        return []

    @contextmanager
    def cleanup_threads(
        event: threading.Event, printer_thread: threading.Thread, executor: concurrent.futures.ThreadPoolExecutor
    ) -> Generator[None]:
        try:
            yield
        except:
            _PRINT_QUEUE.put("Shutting down worker threads...")
            event.set()
            printer_thread.join()
            executor.shutdown(cancel_futures=True)
            raise

    event = threading.Event()
    printer_thread = threading.Thread(target=print_queued_messages, args=(event,))
    printer_thread.start()

    with concurrent.futures.ThreadPoolExecutor(max_workers=os.cpu_count()) as executor:
        # Each temporary directory may be used by multiple processes concurrently during the next step;
        # must make sure that they're all setup correctly before starting the next step,
        # in order to avoid race conditions
        testcase_futures = [
            executor.submit(setup_testcase_dir, package, tempdir, verbosity)
            for package, tempdir in packageinfo_to_tempdir.items()
        ]

        with cleanup_threads(event, printer_thread, executor):
            concurrent.futures.wait(testcase_futures)

        mypy_futures = [executor.submit(task) for task in to_do]

        with cleanup_threads(event, printer_thread, executor):
            results = [future.result() for future in mypy_futures]

    event.set()
    printer_thread.join()
    return results


def main() -> ReturnCode:
    args = parser.parse_args()

    testcase_directories = args.packages_to_test or get_all_testcase_directories()
    verbosity = Verbosity[args.verbosity]
    if args.all:
        if args.platforms_to_test:
            parser.error("Cannot specify both --platform and --all")
        if args.versions_to_test:
            parser.error("Cannot specify both --python-version and --all")
        platforms_to_test, versions_to_test = SUPPORTED_PLATFORMS, SUPPORTED_VERSIONS
    else:
        platforms_to_test = args.platforms_to_test or [sys.platform]
        versions_to_test = args.versions_to_test or [PYTHON_VERSION]

    results: list[Result] | None = None

    with ExitStack() as stack:
        results = concurrently_run_testcases(stack, testcase_directories, verbosity, platforms_to_test, versions_to_test)

    assert results is not None
    if not results:
        print_error("All tests were skipped!")
        return 1

    print()

    for result in results:
        result.print_description()

    code = max(result.code for result in results)

    if code:
        print_error("Test completed with errors")
    else:
        print(colored("Test completed successfully!", "green"))

    return code


if __name__ == "__main__":
    try:
        code = main()
    except KeyboardInterrupt:
        print_error("Test aborted due to KeyboardInterrupt!")
        code = 1
    raise SystemExit(code)



from _typeshed import Incomplete

from braintree.us_bank_account import UsBankAccount

class UsBankAccountGateway:
    gateway: Incomplete
    config: Incomplete
    def __init__(self, gateway) -> None: ...
    def find(self, us_bank_account_token: str) -> UsBankAccount | None: ...



from _typeshed import Incomplete
from collections.abc import Callable, Generator, Iterable

from networkx.classes.digraph import DiGraph
from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = [
    "descendants",
    "ancestors",
    "topological_sort",
    "lexicographical_topological_sort",
    "all_topological_sorts",
    "topological_generations",
    "is_directed_acyclic_graph",
    "is_aperiodic",
    "transitive_closure",
    "transitive_closure_dag",
    "transitive_reduction",
    "antichains",
    "dag_longest_path",
    "dag_longest_path_length",
    "dag_to_branching",
    "compute_v_structures",
]

@_dispatchable
def descendants(G: Graph[_Node], source) -> set[_Node]: ...
@_dispatchable
def ancestors(G: Graph[_Node], source) -> set[_Node]: ...
@_dispatchable
def is_directed_acyclic_graph(G: Graph[_Node]) -> bool: ...
@_dispatchable
def topological_generations(G: DiGraph[_Node]) -> Generator[list[_Node], None, None]: ...
@_dispatchable
def topological_sort(G: DiGraph[_Node]) -> Generator[_Node, None, None]: ...
@_dispatchable
def lexicographical_topological_sort(
    G: DiGraph[_Node], key: Callable[..., Incomplete] | None = None
) -> Generator[_Node, None, None]: ...
@_dispatchable
def all_topological_sorts(G: DiGraph[_Node]) -> Generator[list[_Node], None, None]: ...
@_dispatchable
def is_aperiodic(G: DiGraph[_Node]) -> bool: ...
@_dispatchable
def transitive_closure(G: Graph[_Node], reflexive=False) -> Graph[_Node]: ...
@_dispatchable
def transitive_closure_dag(G: DiGraph[_Node], topo_order: Iterable[Incomplete] | None = None) -> DiGraph[_Node]: ...
@_dispatchable
def transitive_reduction(G: DiGraph[_Node]) -> DiGraph[_Node]: ...
@_dispatchable
def antichains(G: DiGraph[_Node], topo_order: Iterable[Incomplete] | None = None) -> Generator[list[_Node], None, None]: ...
@_dispatchable
def dag_longest_path(
    G: DiGraph[_Node],
    weight: str | None = "weight",
    default_weight: int | None = 1,
    topo_order: Iterable[Incomplete] | None = None,
) -> list[_Node]: ...
@_dispatchable
def dag_longest_path_length(G: DiGraph[_Node], weight: str | None = "weight", default_weight: int | None = 1) -> int: ...
@_dispatchable
def dag_to_branching(G: DiGraph[_Node]) -> DiGraph[_Node]: ...
@_dispatchable
def compute_v_structures(G: DiGraph[_Node]) -> Generator[tuple[Incomplete, Incomplete, Incomplete]]: ...



"""
A small util to simplify the creation of Parsers for simple context-free-grammars.

"""

from abc import ABC, abstractmethod
from collections import defaultdict
from functools import wraps
from types import FunctionType, MethodType
from typing import Generic, TypeVar, Optional, List

__all__ = ['nomatch', 'NoMatch', 'SimpleParser']


class nomatch(BaseException):
    def __init__(self):
        pass


class NoMatch(ValueError):
    def __init__(self, data: str, index: int, expected: List[str]):
        self.data = data
        self.index = index
        self.expected = expected
        super(NoMatch, self).__init__(f"Can not match at index {index}. Got {data[index:index + 5]!r},"
                                      f" expected any of {expected}.\n"
                                      f"Context(data[-10:+10]): {data[index - 10: index + 10]!r}")


T = TypeVar('T')


def _wrap_reset(m):
    @wraps(m)
    def w(self, *args, **kwargs):
        p = self.index
        try:
            return m(self, *args, **kwargs)
        except nomatch:
            self.index = p
            raise

    return w


class SimpleParser(Generic[T], ABC):
    def __init__(self, data: str):
        self.data = data
        self.index = 0
        self._expected = defaultdict(list)

    def __init_subclass__(cls, **kwargs):
        for n, v in cls.__dict__.items():
            if isinstance(v, FunctionType) and not n.startswith('_'):
                setattr(cls, n, _wrap_reset(v))

    def parse(self) -> T:
        try:
            result = self.start()
        except nomatch:
            raise NoMatch(self.data, max(self._expected), self._expected[max(self._expected)]) from None
        if self.index < len(self.data):
            raise NoMatch(self.data, max(self._expected), self._expected[max(self._expected)])
        return result

    @abstractmethod
    def start(self) -> T:
        raise NotImplementedError

    def peek_static(self, expected: str) -> bool:
        l = len(expected)
        if self.data[self.index:self.index + l] == expected:
            return True
        else:
            self._expected[self.index].append(expected)
            return False

    def static(self, expected: str):
        length = len(expected)
        if self.data[self.index:self.index + length] == expected:
            self.index += length
        else:
            self._expected[self.index].append(expected)
            raise nomatch

    def static_b(self, expected: str) -> bool:
        l = len(expected)
        if self.data[self.index:self.index + l] == expected:
            self.index += l
            return True
        else:
            self._expected[self.index].append(expected)
            return False

    def anyof(self, *strings: str) -> str:
        for s in strings:
            if self.static_b(s):
                return s
        else:
            raise nomatch

    def anyof_b(self, *strings: str) -> bool:
        for s in strings:
            if self.static_b(s):
                return True
        else:
            return False

    def any(self, length: int = 1) -> str:
        if self.index + length <= len(self.data):
            res = self.data[self.index:self.index + length]
            self.index += length
            return res
        else:
            self._expected[self.index].append(f"<Any {length}>")
            raise nomatch

    def any_but(self, *strings, length: int = 1) -> str:
        if self.index + length <= len(self.data):
            res = self.data[self.index:self.index + length]
            if res not in strings:
                self.index += length
                return res
            else:
                self._expected[self.index].append(f"<Any {length} except {strings}>")
                raise nomatch
        else:
            self._expected[self.index].append(f"<Any {length} except {strings}>")
            raise nomatch

    def multiple(self, chars: str, mi: int, ma: Optional[int]) -> str:
        result = []
        try:
            for off in range(mi):
                if self.data[self.index + off] in chars:
                    result.append(self.data[self.index + off])
                else:
                    self._expected[self.index + off].extend(chars)
                    raise nomatch
        except IndexError:
            raise nomatch
        self.index += mi
        if ma is None:
            try:
                while True:
                    if self.data[self.index] in chars:
                        result.append(self.data[self.index])
                        self.index += 1
                    else:
                        self._expected[self.index].extend(chars)
                        break
            except IndexError:
                pass
        else:
            try:
                for _ in range(ma - mi):
                    if self.data[self.index] in chars:
                        result.append(self.data[self.index])
                        self.index += 1
                    else:
                        self._expected[self.index].extend(chars)
                        break
            except IndexError:
                pass
        return ''.join(result)



from . import BaseDirectory, Config, DesktopEntry, Exceptions, IconTheme, IniFile, Locale, Menu, MenuEditor, Mime, RecentFiles

__all__ = [
    "BaseDirectory",
    "Config",
    "DesktopEntry",
    "Exceptions",
    "IconTheme",
    "IniFile",
    "Locale",
    "Menu",
    "MenuEditor",
    "Mime",
    "RecentFiles",
]



"""
Tests for `defaultdict.__or__` and `defaultdict.__ror__`.
"""

from __future__ import annotations

import os
from collections import defaultdict
from typing import Mapping, TypeVar, Union
from typing_extensions import Self, assert_type

_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


class CustomDefaultDictSubclass(defaultdict[_KT, _VT]):
    pass


class CustomMappingWithDunderOr(Mapping[_KT, _VT]):
    def __or__(self, other: Mapping[_KT, _VT]) -> dict[_KT, _VT]:
        return {}

    def __ror__(self, other: Mapping[_KT, _VT]) -> dict[_KT, _VT]:
        return {}

    def __ior__(self, other: Mapping[_KT, _VT]) -> Self:
        return self


def test_defaultdict_dot_or(
    a: defaultdict[int, int],
    b: CustomDefaultDictSubclass[int, int],
    c: defaultdict[str, str],
    d: Mapping[int, int],
    e: CustomMappingWithDunderOr[str, str],
) -> None:
    assert_type(a | b, defaultdict[int, int])

    # In contrast to `dict.__or__`, `defaultdict.__or__` returns `Self` if called on a subclass of `defaultdict`:
    assert_type(b | a, CustomDefaultDictSubclass[int, int])

    assert_type(a | c, defaultdict[Union[int, str], Union[int, str]])

    # arbitrary mappings are not accepted by `defaultdict.__or__`;
    # it has to be a subclass of `dict`
    a | d  # type: ignore

    # but Mappings such as `os._Environ` or `CustomMappingWithDunderOr`,
    # which define `__ror__` methods that accept `dict`, are fine
    # (`os._Environ.__(r)or__` always returns `dict`, even if a `defaultdict` is passed):
    assert_type(a | os.environ, dict[Union[str, int], Union[str, int]])
    assert_type(os.environ | a, dict[Union[str, int], Union[str, int]])

    assert_type(c | os.environ, dict[str, str])
    assert_type(c | e, dict[str, str])

    assert_type(os.environ | c, dict[str, str])
    assert_type(e | c, dict[str, str])

    # store "untainted" `CustomMappingWithDunderOr[str, str]` to test `__ior__` against ` defaultdict[str, str]` later
    # Invalid `e |= a` causes pyright to join `Unknown` to `e`'s type
    f = e

    e |= c
    e |= a  # type: ignore

    c |= f

    c |= a  # type: ignore



import sys
from collections import deque
from collections.abc import Mapping as AbcMapping
from collections.abc import MutableMapping as AbcMutableMapping
from collections.abc import MutableSet as AbcMutableSet
from collections.abc import Set as AbcSet
from dataclasses import MISSING, Field, is_dataclass
from dataclasses import fields as dataclass_fields
from functools import partial
from inspect import signature as _signature
from typing import AbstractSet as TypingAbstractSet
from typing import (
    Any,
    Deque,
    Dict,
    Final,
    FrozenSet,
    List,
    Literal,
    NewType,
    Optional,
    Protocol,
    Tuple,
    Type,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)
from typing import Mapping as TypingMapping
from typing import MutableMapping as TypingMutableMapping
from typing import MutableSequence as TypingMutableSequence
from typing import MutableSet as TypingMutableSet
from typing import Sequence as TypingSequence
from typing import Set as TypingSet

from attrs import NOTHING, Attribute, Factory, resolve_types
from attrs import fields as attrs_fields
from attrs import fields_dict as attrs_fields_dict

__all__ = [
    "ANIES",
    "adapted_fields",
    "fields_dict",
    "ExceptionGroup",
    "ExtensionsTypedDict",
    "get_type_alias_base",
    "has",
    "is_type_alias",
    "is_typeddict",
    "TypeAlias",
    "TypedDict",
]

try:
    from typing_extensions import TypedDict as ExtensionsTypedDict
except ImportError:  # pragma: no cover
    ExtensionsTypedDict = None

if sys.version_info >= (3, 11):
    from builtins import ExceptionGroup
else:
    from exceptiongroup import ExceptionGroup

try:
    from typing_extensions import is_typeddict as _is_typeddict
except ImportError:  # pragma: no cover
    assert sys.version_info >= (3, 10)
    from typing import is_typeddict as _is_typeddict

try:
    from typing_extensions import TypeAlias
except ImportError:  # pragma: no cover
    assert sys.version_info >= (3, 11)
    from typing import TypeAlias

LITERALS = {Literal}
try:
    from typing_extensions import Literal as teLiteral

    LITERALS.add(teLiteral)
except ImportError:  # pragma: no cover
    pass

# On some Python versions, `typing_extensions.Any` is different than
# `typing.Any`.
try:
    from typing_extensions import Any as teAny

    ANIES = frozenset([Any, teAny])
except ImportError:  # pragma: no cover
    ANIES = frozenset([Any])

NoneType = type(None)


def is_optional(typ: Type) -> bool:
    return is_union_type(typ) and NoneType in typ.__args__ and len(typ.__args__) == 2


def is_typeddict(cls):
    """Thin wrapper around typing(_extensions).is_typeddict"""
    return _is_typeddict(getattr(cls, "__origin__", cls))


def is_type_alias(type: Any) -> bool:
    """Is this a PEP 695 type alias?"""
    return False


def get_type_alias_base(type: Any) -> Any:
    """
    What is this a type alias of?

    Works only on 3.12+.
    """
    return type.__value__


def has(cls):
    return hasattr(cls, "__attrs_attrs__") or hasattr(cls, "__dataclass_fields__")


def has_with_generic(cls):
    """Test whether the class if a normal or generic attrs or dataclass."""
    return has(cls) or has(get_origin(cls))


def fields(type):
    try:
        return type.__attrs_attrs__
    except AttributeError:
        return dataclass_fields(type)


def fields_dict(type) -> Dict[str, Union[Attribute, Field]]:
    """Return the fields_dict for attrs and dataclasses."""
    if is_dataclass(type):
        return {f.name: f for f in dataclass_fields(type)}
    return attrs_fields_dict(type)


def adapted_fields(cl) -> List[Attribute]:
    """Return the attrs format of `fields()` for attrs and dataclasses."""
    if is_dataclass(cl):
        attrs = dataclass_fields(cl)
        if any(isinstance(a.type, str) for a in attrs):
            # Do this conditionally in case `get_type_hints` fails, so
            # users can resolve on their own first.
            type_hints = get_type_hints(cl)
        else:
            type_hints = {}
        return [
            Attribute(
                attr.name,
                (
                    attr.default
                    if attr.default is not MISSING
                    else (
                        Factory(attr.default_factory)
                        if attr.default_factory is not MISSING
                        else NOTHING
                    )
                ),
                None,
                True,
                None,
                True,
                attr.init,
                True,
                type=type_hints.get(attr.name, attr.type),
                alias=attr.name,
                kw_only=getattr(attr, "kw_only", False),
            )
            for attr in attrs
        ]
    attribs = attrs_fields(cl)
    if any(isinstance(a.type, str) for a in attribs):
        # PEP 563 annotations - need to be resolved.
        resolve_types(cl)
        attribs = attrs_fields(cl)
    return attribs


def is_subclass(obj: type, bases) -> bool:
    """A safe version of issubclass (won't raise)."""
    try:
        return issubclass(obj, bases)
    except TypeError:
        return False


def is_hetero_tuple(type: Any) -> bool:
    origin = getattr(type, "__origin__", None)
    return origin is tuple and ... not in type.__args__


def is_protocol(type: Any) -> bool:
    return is_subclass(type, Protocol) and getattr(type, "_is_protocol", False)


def is_bare_final(type) -> bool:
    return type is Final


def get_final_base(type) -> Optional[type]:
    """Return the base of the Final annotation, if it is Final."""
    if type is Final:
        return Any
    if type.__class__ is _GenericAlias and type.__origin__ is Final:
        return type.__args__[0]
    return None


OriginAbstractSet = AbcSet
OriginMutableSet = AbcMutableSet

signature = _signature

if sys.version_info >= (3, 10):
    signature = partial(_signature, eval_str=True)

if sys.version_info >= (3, 9):
    from collections import Counter
    from collections.abc import MutableSequence as AbcMutableSequence
    from collections.abc import MutableSet as AbcMutableSet
    from collections.abc import Sequence as AbcSequence
    from collections.abc import Set as AbcSet
    from types import GenericAlias
    from typing import (
        Annotated,
        Generic,
        TypedDict,
        Union,
        _AnnotatedAlias,
        _GenericAlias,
        _SpecialGenericAlias,
        _UnionGenericAlias,
    )
    from typing import Counter as TypingCounter

    try:
        # Not present on 3.9.0, so we try carefully.
        from typing import _LiteralGenericAlias

        def is_literal(type) -> bool:
            return type in LITERALS or (
                isinstance(
                    type, (_GenericAlias, _LiteralGenericAlias, _SpecialGenericAlias)
                )
                and type.__origin__ in LITERALS
            )

    except ImportError:  # pragma: no cover

        def is_literal(_) -> bool:
            return False

    Set = AbcSet
    AbstractSet = AbcSet
    MutableSet = AbcMutableSet
    Sequence = AbcSequence
    MutableSequence = AbcMutableSequence
    MutableMapping = AbcMutableMapping
    Mapping = AbcMapping
    FrozenSetSubscriptable = frozenset
    TupleSubscriptable = tuple

    def is_annotated(type) -> bool:
        return getattr(type, "__class__", None) is _AnnotatedAlias

    def is_tuple(type):
        return (
            type in (Tuple, tuple)
            or (type.__class__ is _GenericAlias and is_subclass(type.__origin__, Tuple))
            or (getattr(type, "__origin__", None) is tuple)
        )

    if sys.version_info >= (3, 12):
        from typing import TypeAliasType

        def is_type_alias(type: Any) -> bool:
            """Is this a PEP 695 type alias?"""
            return isinstance(type, TypeAliasType)

    if sys.version_info >= (3, 10):

        def is_union_type(obj):
            from types import UnionType

            return (
                obj is Union
                or (isinstance(obj, _UnionGenericAlias) and obj.__origin__ is Union)
                or isinstance(obj, UnionType)
            )

        def get_newtype_base(typ: Any) -> Optional[type]:
            if typ is NewType or isinstance(typ, NewType):
                return typ.__supertype__
            return None

        if sys.version_info >= (3, 11):
            from typing import NotRequired, Required
        else:
            from typing_extensions import NotRequired, Required

    else:
        from typing_extensions import NotRequired, Required

        def is_union_type(obj):
            return (
                obj is Union
                or isinstance(obj, _UnionGenericAlias)
                and obj.__origin__ is Union
            )

        def get_newtype_base(typ: Any) -> Optional[type]:
            supertype = getattr(typ, "__supertype__", None)
            if (
                supertype is not None
                and getattr(typ, "__qualname__", "") == "NewType.<locals>.new_type"
                and typ.__module__ in ("typing", "typing_extensions")
            ):
                return supertype
            return None

    def get_notrequired_base(type) -> "Union[Any, Literal[NOTHING]]":
        if is_annotated(type):
            # Handle `Annotated[NotRequired[int]]`
            type = get_args(type)[0]
        if get_origin(type) in (NotRequired, Required):
            return get_args(type)[0]
        return NOTHING

    def is_sequence(type: Any) -> bool:
        """A predicate function for sequences.

        Matches lists, sequences, mutable sequences, deques and homogenous
        tuples.
        """
        origin = getattr(type, "__origin__", None)
        return (
            type
            in (
                List,
                list,
                TypingSequence,
                TypingMutableSequence,
                AbcMutableSequence,
                tuple,
                Tuple,
                deque,
                Deque,
            )
            or (
                type.__class__ is _GenericAlias
                and (
                    (origin is not tuple)
                    and is_subclass(origin, TypingSequence)
                    or origin is tuple
                    and type.__args__[1] is ...
                )
            )
            or (origin in (list, deque, AbcMutableSequence, AbcSequence))
            or (origin is tuple and type.__args__[1] is ...)
        )

    def is_deque(type):
        return (
            type in (deque, Deque)
            or (type.__class__ is _GenericAlias and is_subclass(type.__origin__, deque))
            or (getattr(type, "__origin__", None) is deque)
        )

    def is_mutable_set(type: Any) -> bool:
        """A predicate function for (mutable) sets.

        Matches built-in sets and sets from the typing module.
        """
        return (
            type in (TypingSet, TypingMutableSet, set)
            or (
                type.__class__ is _GenericAlias
                and is_subclass(type.__origin__, TypingMutableSet)
            )
            or (getattr(type, "__origin__", None) in (set, AbcMutableSet, AbcSet))
        )

    def is_frozenset(type: Any) -> bool:
        """A predicate function for frozensets.

        Matches built-in frozensets and frozensets from the typing module.
        """
        return (
            type in (FrozenSet, frozenset)
            or (
                type.__class__ is _GenericAlias
                and is_subclass(type.__origin__, FrozenSet)
            )
            or (getattr(type, "__origin__", None) is frozenset)
        )

    def is_bare(type):
        return isinstance(type, _SpecialGenericAlias) or (
            not hasattr(type, "__origin__") and not hasattr(type, "__args__")
        )

    def is_mapping(type: Any) -> bool:
        """A predicate function for mappings."""
        return (
            type in (dict, Dict, TypingMapping, TypingMutableMapping, AbcMutableMapping)
            or (
                type.__class__ is _GenericAlias
                and is_subclass(type.__origin__, TypingMapping)
            )
            or is_subclass(
                getattr(type, "__origin__", type), (dict, AbcMutableMapping, AbcMapping)
            )
        )

    def is_counter(type):
        return (
            type in (Counter, TypingCounter)
            or getattr(type, "__origin__", None) is Counter
        )

    def is_generic(type) -> bool:
        """Whether `type` is a generic type."""
        # Inheriting from protocol will inject `Generic` into the MRO
        # without `__orig_bases__`.
        return isinstance(type, (_GenericAlias, GenericAlias)) or (
            is_subclass(type, Generic) and hasattr(type, "__orig_bases__")
        )

    def copy_with(type, args):
        """Replace a generic type's arguments."""
        if is_annotated(type):
            # typing.Annotated requires a special case.
            return Annotated[args]
        if isinstance(args, tuple) and len(args) == 1:
            # Some annotations can't handle 1-tuples.
            args = args[0]
        return type.__origin__[args]

    def get_full_type_hints(obj, globalns=None, localns=None):
        return get_type_hints(obj, globalns, localns, include_extras=True)

else:
    # 3.8
    Set = TypingSet
    AbstractSet = TypingAbstractSet
    MutableSet = TypingMutableSet

    Sequence = TypingSequence
    MutableSequence = TypingMutableSequence
    MutableMapping = TypingMutableMapping
    Mapping = TypingMapping
    FrozenSetSubscriptable = FrozenSet
    TupleSubscriptable = Tuple

    from collections import Counter as ColCounter
    from typing import Counter, Generic, TypedDict, Union, _GenericAlias

    from typing_extensions import Annotated, NotRequired, Required
    from typing_extensions import get_origin as te_get_origin

    def is_annotated(type) -> bool:
        return te_get_origin(type) is Annotated

    def is_tuple(type):
        return type in (Tuple, tuple) or (
            type.__class__ is _GenericAlias and is_subclass(type.__origin__, Tuple)
        )

    def is_union_type(obj):
        return (
            obj is Union or isinstance(obj, _GenericAlias) and obj.__origin__ is Union
        )

    def get_newtype_base(typ: Any) -> Optional[type]:
        supertype = getattr(typ, "__supertype__", None)
        if (
            supertype is not None
            and getattr(typ, "__qualname__", "") == "NewType.<locals>.new_type"
            and typ.__module__ in ("typing", "typing_extensions")
        ):
            return supertype
        return None

    def is_sequence(type: Any) -> bool:
        return type in (List, list, Tuple, tuple) or (
            type.__class__ is _GenericAlias
            and (
                type.__origin__ not in (Union, Tuple, tuple)
                and is_subclass(type.__origin__, TypingSequence)
            )
            or (type.__origin__ in (Tuple, tuple) and type.__args__[1] is ...)
        )

    def is_deque(type: Any) -> bool:
        return (
            type in (deque, Deque)
            or (type.__class__ is _GenericAlias and is_subclass(type.__origin__, deque))
            or type.__origin__ is deque
        )

    def is_mutable_set(type) -> bool:
        return type in (set, TypingAbstractSet) or (
            type.__class__ is _GenericAlias
            and is_subclass(type.__origin__, (MutableSet, TypingAbstractSet))
        )

    def is_frozenset(type):
        return type is frozenset or (
            type.__class__ is _GenericAlias and is_subclass(type.__origin__, FrozenSet)
        )

    def is_mapping(type: Any) -> bool:
        """A predicate function for mappings."""
        return (
            type in (TypingMapping, dict)
            or (
                type.__class__ is _GenericAlias
                and is_subclass(type.__origin__, TypingMapping)
            )
            or is_subclass(
                getattr(type, "__origin__", type), (dict, AbcMutableMapping, AbcMapping)
            )
        )

    bare_generic_args = {
        List.__args__,
        TypingSequence.__args__,
        TypingMapping.__args__,
        Dict.__args__,
        TypingMutableSequence.__args__,
        Tuple.__args__,
        None,  # non-parametrized containers do not have `__args__ attribute in py3.7-8
    }

    def is_bare(type):
        return getattr(type, "__args__", None) in bare_generic_args

    def is_counter(type):
        return (
            type in (Counter, ColCounter)
            or getattr(type, "__origin__", None) is ColCounter
        )

    def is_literal(type) -> bool:
        return type in LITERALS or (
            isinstance(type, _GenericAlias) and type.__origin__ in LITERALS
        )

    def is_generic(obj):
        return isinstance(obj, _GenericAlias) or (
            is_subclass(obj, Generic) and hasattr(obj, "__orig_bases__")
        )

    def copy_with(type, args):
        """Replace a generic type's arguments."""
        return type.copy_with(args)

    def get_notrequired_base(type) -> "Union[Any, Literal[NOTHING]]":
        if is_annotated(type):
            # Handle `Annotated[NotRequired[int]]`
            type = get_origin(type)

        if get_origin(type) in (NotRequired, Required):
            return get_args(type)[0]
        return NOTHING

    def get_full_type_hints(obj, globalns=None, localns=None):
        return get_type_hints(obj, globalns, localns)


def is_generic_attrs(type) -> bool:
    """Return True for both specialized (A[int]) and unspecialized (A) generics."""
    return is_generic(type) and has(type.__origin__)



from collections.abc import Callable
from typing import Final

# Gleaned from https://github.com/google/jsonnet/blob/master/python/_jsonnet.c
version: Final[str]

def evaluate_file(
    filename: str,
    jpathdir: str | list[str] | None = ...,
    max_stack: int = 500,
    gc_min_objects: int = 1000,
    gc_growth_trigger: float = 2,
    ext_vars: dict[str, str] | None = ...,
    ext_codes: dict[str, str] | None = ...,
    tla_vars: dict[str, str] | None = ...,
    tla_codes: dict[str, str] | None = ...,
    max_trace: int = 20,
    import_callback: Callable[[str, str], tuple[str, object | None]] = ...,
    native_callbacks: dict[str, tuple[tuple[str, ...], Callable[..., object]]] | None = ...,
) -> str: ...
def evaluate_snippet(
    filename: str,
    src: str,
    jpathdir: str | list[str] | None = ...,
    max_stack: int = 500,
    gc_min_objects: int = 1000,
    gc_growth_trigger: float = 2,
    ext_vars: dict[str, str] | None = ...,
    ext_codes: dict[str, str] | None = ...,
    tla_vars: dict[str, str] | None = ...,
    tla_codes: dict[str, str] | None = ...,
    max_trace: int = 20,
    import_callback: Callable[[str, str], tuple[str, object | None]] = ...,
    native_callbacks: dict[str, tuple[tuple[str, ...], Callable[..., object]]] | None = ...,
) -> str: ...



from collections.abc import Callable
from io import BytesIO
from typing import Any

python_version: tuple[int, int]

BYTES_LITERAL: Callable[[str], bytes]
UNICODE_LITERAL: Callable[[str], str]
BYTES_ORD: Callable[[bytes], int]
BYTES_IO: type[BytesIO]

def fprintf(f: Any, fmt: str, *vargs: Any) -> None: ...

EXCEL_TEXT_TYPES: tuple[type[str], type[bytes], type[bytearray]]
REPR = ascii
xrange = range
unicode: Callable[[bytes, str], str]
ensure_unicode: Callable[[str | bytes], str]
unichr = chr



from _typeshed import StrPath
from collections.abc import Callable
from typing import Any

from ..extractor.common import _InfoDict
from ..YoutubeDL import YoutubeDL

class PostProcessorMetaClass(type):
    @staticmethod
    def run_wrapper(func: Callable[..., object]) -> Callable[..., object]: ...
    def __new__(cls, name: str, bases: tuple[type[Any], ...], attrs: dict[str, Any]) -> type[Any]: ...

class PostProcessor(metaclass=PostProcessorMetaClass):
    PP_NAME: str
    def __init__(self, downloader: YoutubeDL | None = None) -> None: ...
    @classmethod
    def pp_key(cls) -> str: ...
    def to_screen(
        self,
        text: str,
        prefix: bool = True,
        *,
        message: str,
        skip_eol: bool = False,
        quiet: bool | None = None,
        only_once: bool = False,
    ) -> None: ...
    def report_warning(self, text: str, only_once: bool = False) -> None: ...
    def deprecation_warning(self, msg: str) -> None: ...
    def deprecated_feature(self, msg: str) -> None: ...
    def write_debug(self, text: str, *, only_once: bool = False) -> None: ...
    # *args and **kwargs are passed to .param.get() where param is normally a dict but does not have to be.
    def get_param(self, name: str, default: Any = None, *args: Any, **kwargs: Any) -> Any: ...
    def set_downloader(self, downloader: YoutubeDL) -> None: ...
    def run(self, information: _InfoDict) -> tuple[list[str], _InfoDict]: ...
    def try_utime(self, path: StrPath, atime: int, mtime: int, errnote: str = "Cannot update utime of file") -> None: ...
    def add_progress_hook(self, ph: Callable[[str], object]) -> None: ...
    def report_progress(self, s: str) -> None: ...



# pyright: reportWildcardImportFromLibrary=false
"""
This tests that star imports work when using "all += " syntax.
"""
from __future__ import annotations

from typing import *
from zipfile import *

x: Annotated[int, 42]

p: Path



from abc import ABC
from dataclasses import dataclass
from typing import Any
from typing_extensions import Self

from .common import Request, RequestHandler

@dataclass(order=True, frozen=True)
class ImpersonateTarget:
    client: str | None = ...
    version: str | None = ...
    os: str | None = ...
    os_version: str | None = ...

    def __post_init__(self) -> None: ...
    def __contains__(self, target: Self) -> bool: ...
    @classmethod
    def from_str(cls, target: str) -> Self: ...

class ImpersonateRequestHandler(RequestHandler, ABC):
    _SUPPORTED_IMPERSONATE_TARGET_MAP: dict[ImpersonateTarget, Any]  # Copied from source.

    def __init__(
        self,
        *,
        impersonate: ImpersonateTarget | None = None,
        # All keyword arguments are ignored (passed to RequestHandler as **kwargs but RequestHandler.__init__() has **_ and does
        # not use it).
        **kwargs: Any,
    ) -> None: ...
    @property
    def supported_targets(cls) -> tuple[ImpersonateTarget, ...]: ...
    def is_supported_target(self, target: ImpersonateTarget) -> bool: ...

def impersonate_preference(rh: RequestHandler, request: Request) -> int: ...



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import typing

import google.protobuf.descriptor
import google.protobuf.message

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class PaddingMap(google.protobuf.message.Message):
    """A mapping between the dynamic shape dimension of an input and the arg that
    represents the real shape.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ARG_INDEX_FIELD_NUMBER: builtins.int
    SHAPE_INDEX_FIELD_NUMBER: builtins.int
    PADDING_ARG_INDEX_FIELD_NUMBER: builtins.int
    arg_index: builtins.int
    """Input arg index with dynamic shapes."""
    shape_index: builtins.int
    """The dynamic shape dimension index."""
    padding_arg_index: builtins.int
    """The arg index that dynamic dimension maps to, which represents the value
    of the real shape.
    """
    def __init__(
        self,
        *,
        arg_index: builtins.int | None = ...,
        shape_index: builtins.int | None = ...,
        padding_arg_index: builtins.int | None = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "arg_index", b"arg_index", "padding_arg_index", b"padding_arg_index", "shape_index", b"shape_index"
        ],
    ) -> None: ...

global___PaddingMap = PaddingMap



from _typeshed import Incomplete

class ATNDeserializationOptions:
    __slots__ = ("readonly", "verifyATN", "generateRuleBypassTransitions")
    defaultOptions: Incomplete
    readonly: bool
    verifyATN: Incomplete
    generateRuleBypassTransitions: Incomplete
    def __init__(self, copyFrom: ATNDeserializationOptions | None = None) -> None: ...
    def __setattr__(self, key, value) -> None: ...



from _typeshed import Incomplete
from collections.abc import Collection
from typing import Any, Literal

from networkx._typing import Array2D
from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable
from numpy.typing import DTypeLike
from scipy.sparse import lil_array  # type: ignore[import-untyped]  # pyright: ignore[reportMissingImports]

__all__ = ["attr_matrix", "attr_sparse_matrix"]

@_dispatchable
def attr_matrix(
    G: Graph[_Node],
    edge_attr: str | None = None,
    node_attr: str | None = None,  # runtime also accepts `Callable[[_Node], object]`, but it is not documented
    normalized: bool = False,  # runtime also accepts `Callable[[_Node, _Node], object]`, but it is not documented
    rc_order: Collection[_Node] | None = None,
    dtype: DTypeLike | None = None,
    order: Literal["C", "F"] | None = None,
    # TODO: overload on rc_order and node_attr
    # (rc_order:[node], node_attr:None) -> 2D-array
    # (rc_order:[any], node_attr:str) -> 2D-array
    # (rc_order:None, node_attr:None) -> (2D-array, list[node])
    # (rc_order:None, node_attr:str) -> (2D-array, list[any])
) -> Array2D[Incomplete] | tuple[Array2D[Incomplete], list[_Node] | list[Any]]: ...
@_dispatchable
def attr_sparse_matrix(
    G: Graph[_Node],
    edge_attr: str | None = None,
    node_attr: str | None = None,  # runtime also accepts `Callable[[_Node], object]`, but it is not documented
    normalized: bool = False,  # runtime also accepts `Callable[[_Node, _Node], object]`, but it is not documented
    rc_order: Collection[_Node] | None = None,
    dtype: DTypeLike | None = None,
    # TODO: overload on rc_order and node_attr
    # (rc_order:[node], node_attr:None) -> lil_array
    # (rc_order:[any], node_attr:str) -> lil_array
    # (rc_order:None, node_attr:None) -> (lil_array, list[node])
    # (rc_order:None, node_attr:str) -> (lil_array, list[any])
) -> lil_array | tuple[lil_array, list[_Node] | list[Any]]: ...



from _typeshed import SupportsWrite
from collections.abc import Callable
from types import TracebackType
from typing import Any, Generic, TypeVar
from typing_extensions import ParamSpec, Self

from gevent.hub import Hub
from greenlet import greenlet as greenlet_t

_T = TypeVar("_T")
_P = ParamSpec("_P")

class wrap_errors(Generic[_P, _T]):
    def __init__(self, errors: tuple[type[BaseException], ...], func: Callable[_P, _T]) -> None: ...
    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> _T: ...
    def __getattr__(self, name: str) -> Any: ...

def print_run_info(
    thread_stacks: bool = True, greenlet_stacks: bool = True, limit: int | None = ..., file: SupportsWrite[str] | None = None
) -> None: ...
def format_run_info(
    thread_stacks: bool = True, greenlet_stacks: bool = True, limit: int | None = ..., current_thread_ident: int | None = None
) -> None: ...

class GreenletTree:
    greenlet: greenlet_t | None
    is_current_tree: bool
    child_trees: list[GreenletTree]
    DEFAULT_DETAILS: dict[str, Any]
    def __init__(self, greenlet: greenlet_t | None) -> None: ...
    def add_child(self, tree: GreenletTree) -> None: ...
    @property
    def root(self) -> bool: ...
    def __getattr__(self, name: str) -> Any: ...
    def format_lines(self, details: bool | dict[str, Any] = True) -> str: ...
    def format(self, details: bool | dict[str, Any] = True) -> str: ...
    @classmethod
    def forest(cls) -> list[GreenletTree]: ...
    @classmethod
    def current_tree(cls) -> GreenletTree: ...

class assert_switches:
    hub: Hub | None
    tracer: object | None
    max_blocking_time: float | None
    hub_only: bool
    def __init__(self, max_blocking_time: float | None = None, hub_only: bool = False) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(self, t: type[BaseException] | None, v: BaseException | None, tb: TracebackType | None) -> None: ...

__all__ = ["format_run_info", "print_run_info", "GreenletTree", "wrap_errors", "assert_switches"]



import datetime
from typing import ClassVar

class FairHoliday:
    include_fair_holiday: ClassVar[bool]
    fair_holiday_label: ClassVar[str]

class FairHolidayLastMondayJune(FairHoliday):
    def get_fair_holiday(self, year: int) -> tuple[datetime.date, str]: ...

class FairHolidayFirstMondayJuly(FairHoliday):
    def get_fair_holiday(self, year: int) -> tuple[datetime.date, str]: ...

class FairHolidaySecondMondayJuly(FairHoliday):
    def get_fair_holiday(self, year: int) -> tuple[datetime.date, str]: ...

class FairHolidayThirdMondayJuly(FairHoliday):
    def get_fair_holiday(self, year: int) -> tuple[datetime.date, str]: ...

class FairHolidayLastMondayJuly(FairHoliday):
    def get_fair_holiday(self, year: int) -> tuple[datetime.date, str]: ...

class FairHolidayFourthFridayJuly(FairHoliday):
    def get_fair_holiday(self, year: int) -> tuple[datetime.date, str]: ...

class FairHolidayFirstMondayAugust(FairHoliday):
    def get_fair_holiday(self, year: int) -> tuple[datetime.date, str]: ...



from typing import Final

from .book import Book, Name
from .timemachine import *

__all__ = [
    "oBOOL",
    "oERR",
    "oNUM",
    "oREF",
    "oREL",
    "oSTRG",
    "oUNK",
    "decompile_formula",
    "dump_formula",
    "evaluate_name_formula",
    "okind_dict",
    "rangename3d",
    "rangename3drel",
    "cellname",
    "cellnameabs",
    "colname",
    "FMLA_TYPE_CELL",
    "FMLA_TYPE_SHARED",
    "FMLA_TYPE_ARRAY",
    "FMLA_TYPE_COND_FMT",
    "FMLA_TYPE_DATA_VAL",
    "FMLA_TYPE_NAME",
    "Operand",
    "Ref3D",
]

FMLA_TYPE_CELL: Final[int]
FMLA_TYPE_SHARED: Final[int]
FMLA_TYPE_ARRAY: Final[int]
FMLA_TYPE_COND_FMT: Final[int]
FMLA_TYPE_DATA_VAL: Final[int]
FMLA_TYPE_NAME: Final[int]
oBOOL: Final[int]
oERR: Final[int]
oNUM: Final[int]
oREF: Final[int]
oREL: Final[int]
oSTRG: Final[int]
oUNK: Final[int]
okind_dict: Final[dict[int, str]]

class FormulaError(Exception): ...

class Operand:
    value: float | str | None
    kind: int
    text: str
    rank: int
    def __init__(self, akind: int | None = None, avalue: float | str | None = None, arank: int = 0, atext: str = "?") -> None: ...

class Ref3D(tuple[int, int, int, int, int, int, int, int, int, int, int, int]):
    coords: tuple[int, int, int, int, int, int]
    relflags: tuple[int, int, int, int, int, int]
    shtxlo: int
    shtxhi: int
    rowxlo: int
    rowxhi: int
    colxlo: int
    colxhi: int
    def __init__(  # pyright: ignore[reportInconsistentConstructor]
        self, atuple: tuple[int, int, int, int, int, int, int, int, int, int, int, int]
    ) -> None: ...

def evaluate_name_formula(bk: Book, nobj: Name, namex: str, blah: int = 0, level: int = 0) -> None: ...
def decompile_formula(
    bk: Book,
    fmla: bytes,
    fmlalen: int,
    fmlatype: int | None = None,
    browx: int | None = None,
    bcolx: int | None = None,
    blah: int = 0,
    level: int = 0,
    r1c1: int = 0,
) -> str | None: ...
def dump_formula(bk: Book, data: bytes, fmlalen: int, bv: int, reldelta: int, blah: int = 0, isname: int = 0) -> None: ...
def cellname(rowx: int, colx: int) -> str: ...
def cellnameabs(rowx: int, colx: int, r1c1: int = 0) -> str: ...
def colname(colx: int) -> str: ...
def rangename3d(book: Book, ref3d: Ref3D) -> str: ...
def rangename3drel(book: Book, ref3d: Ref3D, browx: int | None = None, bcolx: int | None = None, r1c1: int = 0) -> str: ...



from _typeshed import Incomplete
from collections.abc import Mapping

from authlib.jose import BaseClaims

class ClientMetadataClaims(BaseClaims):
    def validate(self) -> None: ...
    # The "cls" argument is called "self" in the actual implementation,
    # but stubtest will not allow that.
    @classmethod
    def get_claims_options(cls, metadata: Mapping[str, Incomplete]) -> dict[str, Incomplete]: ...
    def validate_token_endpoint_auth_signing_alg(self) -> None: ...
    def validate_application_type(self) -> None: ...
    def validate_sector_identifier_uri(self) -> None: ...
    def validate_subject_type(self) -> None: ...
    def validate_id_token_signed_response_alg(self) -> None: ...
    def validate_id_token_encrypted_response_alg(self) -> None: ...
    def validate_id_token_encrypted_response_enc(self) -> None: ...
    def validate_userinfo_signed_response_alg(self) -> None: ...
    def validate_userinfo_encrypted_response_alg(self) -> None: ...
    def validate_userinfo_encrypted_response_enc(self) -> None: ...
    def validate_default_max_age(self) -> None: ...
    def validate_require_auth_time(self) -> None: ...
    def validate_default_acr_values(self) -> None: ...
    def validate_initiate_login_uri(self) -> None: ...
    def validate_request_object_signing_alg(self) -> None: ...
    def validate_request_object_encryption_alg(self) -> None: ...
    def validate_request_object_encryption_enc(self) -> None: ...
    def validate_request_uris(self) -> None: ...



from collections.abc import Collection
from datetime import datetime
from typing import Literal
from typing_extensions import TypeAlias

from webob._types import AsymmetricPropertyWithDelete
from webob.response import Response

__all__ = ["AnyETag", "NoETag", "ETagMatcher", "IfRange", "etag_property"]

_ETag: TypeAlias = _AnyETag | _NoETag | ETagMatcher
_ETagProperty: TypeAlias = AsymmetricPropertyWithDelete[_ETag, _ETag | str | None]

def etag_property(key: str, default: _ETag, rfc_section: str, strong: bool = True) -> _ETagProperty: ...

class _AnyETag:
    def __bool__(self) -> Literal[False]: ...
    def __contains__(self, other: str | None) -> Literal[True]: ...

AnyETag: _AnyETag

class _NoETag:
    def __bool__(self) -> Literal[False]: ...
    def __contains__(self, other: str | None) -> Literal[False]: ...

NoETag: _NoETag

class ETagMatcher:
    etags: Collection[str]
    def __init__(self, etags: Collection[str]) -> None: ...
    def __contains__(self, other: str | None) -> bool: ...
    @classmethod
    def parse(cls, value: str, strong: bool = True) -> ETagMatcher | _AnyETag: ...

class IfRange:
    etag: _ETag
    def __init__(self, etag: _ETag) -> None: ...
    @classmethod
    def parse(cls, value: str | None) -> IfRange | IfRangeDate: ...
    def __contains__(self, resp: Response) -> bool: ...
    def __bool__(self) -> bool: ...

class IfRangeDate:
    date: datetime
    def __init__(self, date: datetime) -> None: ...
    def __contains__(self, resp: Response) -> bool: ...



from typing import Any
from typing_extensions import TypeAlias

from django.db.models import QuerySet
from django.http import HttpRequest
from django_filters.filterset import FilterSetMetaclass

from . import filterset

# APIView placeholder - djangorestframework is optional, so we use Any for compatibility
_APIView: TypeAlias = Any

class DjangoFilterBackend:
    filterset_base: FilterSetMetaclass = ...
    raise_exception: bool
    @property
    def template(self) -> str: ...

    # Works with any model type
    def get_filterset(self, request: HttpRequest, queryset: QuerySet[Any], view: _APIView) -> filterset.FilterSet | None: ...

    # Any model queryset
    def get_filterset_class(self, view: _APIView, queryset: QuerySet[Any] | None = None) -> type[filterset.FilterSet] | None: ...

    # Kwargs vary by filterset
    def get_filterset_kwargs(self, request: HttpRequest, queryset: QuerySet[Any], view: _APIView) -> dict[str, Any]: ...

    # Filters any model type
    def filter_queryset(self, request: HttpRequest, queryset: QuerySet[Any], view: _APIView) -> QuerySet[Any]: ...
    def to_html(self, request: HttpRequest, queryset: QuerySet[Any], view: _APIView) -> str: ...  # Renders form for any model



import sys
from typing import Any, Literal, overload
from typing_extensions import TypeAlias, TypeGuard

import numpy as np
from numpy.typing import NDArray

from ._typing import ArrayLike, ArrayLikeSeq, OptGeoArrayLike, OptGeoArrayLikeSeq
from .geometry.base import BaseGeometry
from .lib import Geometry

if sys.version_info >= (3, 10):
    _NPTrue: TypeAlias = np.bool_[Literal[True]]
    _NPFalse: TypeAlias = np.bool_[Literal[False]]
else:
    _NPTrue: TypeAlias = np.bool_
    _NPFalse: TypeAlias = np.bool_

__all__ = [
    "contains",
    "contains_properly",
    "contains_xy",
    "covered_by",
    "covers",
    "crosses",
    "disjoint",
    "dwithin",
    "equals",
    "equals_exact",
    "equals_identical",
    "has_m",
    "has_z",
    "intersects",
    "intersects_xy",
    "is_ccw",
    "is_closed",
    "is_empty",
    "is_geometry",
    "is_missing",
    "is_prepared",
    "is_ring",
    "is_simple",
    "is_valid",
    "is_valid_input",
    "is_valid_reason",
    "overlaps",
    "relate",
    "relate_pattern",
    "touches",
    "within",
]

@overload
def has_z(geometry: Geometry | None, **kwargs) -> np.bool_: ...
@overload
def has_z(geometry: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.bool_]: ...
@overload
def has_m(geometry: Geometry | None, **kwargs) -> np.bool_: ...
@overload
def has_m(geometry: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.bool_]: ...
@overload
def is_ccw(geometry: Geometry | None, **kwargs) -> np.bool_: ...
@overload
def is_ccw(geometry: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.bool_]: ...
@overload
def is_closed(geometry: Geometry | None, **kwargs) -> np.bool_: ...
@overload
def is_closed(geometry: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.bool_]: ...
@overload
def is_empty(geometry: Geometry | None, **kwargs) -> np.bool_: ...
@overload
def is_empty(geometry: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.bool_]: ...
@overload
def is_geometry(geometry: Geometry, **kwargs) -> _NPTrue: ...
@overload
def is_geometry(geometry: None, **kwargs) -> _NPFalse: ...
@overload
def is_geometry(geometry: ArrayLikeSeq[Any], **kwargs) -> NDArray[np.bool_]: ...  # type: ignore[overload-overlap]
@overload
def is_geometry(geometry: object, **kwargs) -> TypeGuard[BaseGeometry]: ...
@overload
def is_missing(geometry: Geometry, **kwargs) -> _NPFalse: ...
@overload
def is_missing(geometry: None, **kwargs) -> _NPTrue: ...
@overload
def is_missing(geometry: ArrayLikeSeq[Any], **kwargs) -> NDArray[np.bool_]: ...  # type: ignore[overload-overlap]
@overload
def is_missing(geometry: object, **kwargs) -> TypeGuard[None]: ...
@overload
def is_prepared(geometry: Geometry | None, **kwargs) -> np.bool_: ...
@overload
def is_prepared(geometry: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.bool_]: ...
@overload
def is_valid_input(geometry: Geometry | None, **kwargs) -> _NPTrue: ...
@overload
def is_valid_input(geometry: ArrayLikeSeq[Any], **kwargs) -> NDArray[np.bool_]: ...  # type: ignore[overload-overlap]
@overload
def is_valid_input(geometry: object, **kwargs) -> TypeGuard[BaseGeometry | None]: ...
@overload
def is_ring(geometry: Geometry | None, **kwargs) -> np.bool_: ...
@overload
def is_ring(geometry: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.bool_]: ...
@overload
def is_simple(geometry: Geometry | None, **kwargs) -> np.bool_: ...
@overload
def is_simple(geometry: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.bool_]: ...
@overload
def is_valid(geometry: Geometry | None, **kwargs) -> np.bool_: ...
@overload
def is_valid(geometry: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.bool_]: ...
@overload
def is_valid_reason(geometry: None, **kwargs) -> None: ...
@overload
def is_valid_reason(geometry: Geometry, **kwargs) -> str: ...
@overload
def is_valid_reason(geometry: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.object_]: ...
@overload
def crosses(a: Geometry | None, b: Geometry | None, **kwargs) -> np.bool_: ...
@overload
def crosses(a: OptGeoArrayLikeSeq, b: OptGeoArrayLike, **kwargs) -> NDArray[np.bool_]: ...
@overload
def crosses(a: OptGeoArrayLike, b: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.bool_]: ...
@overload
def contains(a: Geometry | None, b: Geometry | None, **kwargs) -> np.bool_: ...
@overload
def contains(a: OptGeoArrayLikeSeq, b: OptGeoArrayLike, **kwargs) -> NDArray[np.bool_]: ...
@overload
def contains(a: OptGeoArrayLike, b: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.bool_]: ...
@overload
def contains_properly(a: Geometry | None, b: Geometry | None, **kwargs) -> np.bool_: ...
@overload
def contains_properly(a: OptGeoArrayLikeSeq, b: OptGeoArrayLike, **kwargs) -> NDArray[np.bool_]: ...
@overload
def contains_properly(a: OptGeoArrayLike, b: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.bool_]: ...
@overload
def covered_by(a: Geometry | None, b: Geometry | None, **kwargs) -> np.bool_: ...
@overload
def covered_by(a: OptGeoArrayLikeSeq, b: OptGeoArrayLike, **kwargs) -> NDArray[np.bool_]: ...
@overload
def covered_by(a: OptGeoArrayLike, b: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.bool_]: ...
@overload
def covers(a: Geometry | None, b: Geometry | None, **kwargs) -> np.bool_: ...
@overload
def covers(a: OptGeoArrayLikeSeq, b: OptGeoArrayLike, **kwargs) -> NDArray[np.bool_]: ...
@overload
def covers(a: OptGeoArrayLike, b: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.bool_]: ...
@overload
def disjoint(a: Geometry | None, b: Geometry | None, **kwargs) -> np.bool_: ...
@overload
def disjoint(a: OptGeoArrayLikeSeq, b: OptGeoArrayLike, **kwargs) -> NDArray[np.bool_]: ...
@overload
def disjoint(a: OptGeoArrayLike, b: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.bool_]: ...
@overload
def equals(a: Geometry | None, b: Geometry | None, **kwargs) -> np.bool_: ...
@overload
def equals(a: OptGeoArrayLikeSeq, b: OptGeoArrayLike, **kwargs) -> NDArray[np.bool_]: ...
@overload
def equals(a: OptGeoArrayLike, b: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.bool_]: ...
@overload
def intersects(a: Geometry | None, b: Geometry | None, **kwargs) -> np.bool_: ...
@overload
def intersects(a: OptGeoArrayLikeSeq, b: OptGeoArrayLike, **kwargs) -> NDArray[np.bool_]: ...
@overload
def intersects(a: OptGeoArrayLike, b: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.bool_]: ...
@overload
def overlaps(a: Geometry | None, b: Geometry | None, **kwargs) -> np.bool_: ...
@overload
def overlaps(a: OptGeoArrayLikeSeq, b: OptGeoArrayLike, **kwargs) -> NDArray[np.bool_]: ...
@overload
def overlaps(a: OptGeoArrayLike, b: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.bool_]: ...
@overload
def touches(a: Geometry | None, b: Geometry | None, **kwargs) -> np.bool_: ...
@overload
def touches(a: OptGeoArrayLikeSeq, b: OptGeoArrayLike, **kwargs) -> NDArray[np.bool_]: ...
@overload
def touches(a: OptGeoArrayLike, b: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.bool_]: ...
@overload
def within(a: Geometry | None, b: Geometry | None, **kwargs) -> np.bool_: ...
@overload
def within(a: OptGeoArrayLikeSeq, b: OptGeoArrayLike, **kwargs) -> NDArray[np.bool_]: ...
@overload
def within(a: OptGeoArrayLike, b: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.bool_]: ...
@overload
def equals_exact(
    a: Geometry | None, b: Geometry | None, tolerance: float = 0.0, *, normalize: bool = False, **kwargs
) -> np.bool_: ...
@overload
def equals_exact(
    a: OptGeoArrayLike, b: OptGeoArrayLike, tolerance: ArrayLikeSeq[float], *, normalize: bool = False, **kwargs
) -> NDArray[np.bool_]: ...
@overload
def equals_exact(
    a: OptGeoArrayLikeSeq, b: OptGeoArrayLike, tolerance: ArrayLike[float] = 0.0, *, normalize: bool = False, **kwargs
) -> NDArray[np.bool_]: ...
@overload
def equals_exact(
    a: OptGeoArrayLike, b: OptGeoArrayLikeSeq, tolerance: ArrayLike[float] = 0.0, *, normalize: bool = False, **kwargs
) -> NDArray[np.bool_]: ...
@overload
def equals_identical(a: Geometry | None, b: Geometry | None, **kwargs) -> np.bool_: ...
@overload
def equals_identical(a: OptGeoArrayLikeSeq, b: OptGeoArrayLike, **kwargs) -> NDArray[np.bool_]: ...
@overload
def equals_identical(a: OptGeoArrayLike, b: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.bool_]: ...
@overload
def relate(a: Geometry | None, b: None, **kwargs) -> None: ...
@overload
def relate(a: None, b: Geometry | None, **kwargs) -> None: ...
@overload
def relate(a: Geometry, b: Geometry, **kwargs) -> str: ...
@overload
def relate(a: OptGeoArrayLikeSeq, b: OptGeoArrayLike, **kwargs) -> NDArray[np.object_]: ...
@overload
def relate(a: OptGeoArrayLike, b: OptGeoArrayLikeSeq, **kwargs) -> NDArray[np.object_]: ...
@overload
def relate_pattern(a: Geometry | None, b: Geometry | None, pattern: str, **kwargs) -> np.bool_: ...
@overload
def relate_pattern(a: OptGeoArrayLikeSeq, b: OptGeoArrayLike, pattern: str, **kwargs) -> NDArray[np.bool_]: ...
@overload
def relate_pattern(a: OptGeoArrayLike, b: OptGeoArrayLikeSeq, pattern: str, **kwargs) -> NDArray[np.bool_]: ...
@overload
def dwithin(a: Geometry | None, b: Geometry | None, distance: float, **kwargs) -> np.bool_: ...
@overload
def dwithin(a: OptGeoArrayLikeSeq, b: OptGeoArrayLike, distance: float, **kwargs) -> NDArray[np.bool_]: ...
@overload
def dwithin(a: OptGeoArrayLike, b: OptGeoArrayLikeSeq, distance: float, **kwargs) -> NDArray[np.bool_]: ...
@overload
def contains_xy(geom: Geometry | None, x: float, y: float, **kwargs) -> np.bool_: ...
@overload
def contains_xy(geom: OptGeoArrayLike, x: ArrayLikeSeq[float], y: None = None, **kwargs) -> NDArray[np.bool_]: ...
@overload
def contains_xy(geom: Geometry | None, x: ArrayLike[float], y: ArrayLikeSeq[float], **kwargs) -> NDArray[np.bool_]: ...
@overload
def contains_xy(geom: Geometry | None, x: ArrayLikeSeq[float], y: ArrayLike[float], **kwargs) -> NDArray[np.bool_]: ...
@overload
def contains_xy(geom: OptGeoArrayLikeSeq, x: ArrayLike[float], y: ArrayLike[float], **kwargs) -> NDArray[np.bool_]: ...
@overload
def intersects_xy(geom: Geometry | None, x: float, y: float, **kwargs) -> np.bool_: ...
@overload
def intersects_xy(geom: OptGeoArrayLike, x: ArrayLikeSeq[float], y: None = None, **kwargs) -> NDArray[np.bool_]: ...
@overload
def intersects_xy(geom: Geometry | None, x: ArrayLike[float], y: ArrayLikeSeq[float], **kwargs) -> NDArray[np.bool_]: ...
@overload
def intersects_xy(geom: Geometry | None, x: ArrayLikeSeq[float], y: ArrayLike[float], **kwargs) -> NDArray[np.bool_]: ...
@overload
def intersects_xy(geom: OptGeoArrayLikeSeq, x: ArrayLike[float], y: ArrayLike[float], **kwargs) -> NDArray[np.bool_]: ...



from braintree.search import Search

class DisputeSearch:
    amount_disputed: Search.RangeNodeBuilder
    amount_won: Search.RangeNodeBuilder
    case_number: Search.TextNodeBuilder
    chargeback_protection_level: Search.MultipleValueNodeBuilder
    protection_level: Search.MultipleValueNodeBuilder
    customer_id: Search.TextNodeBuilder
    disbursement_date: Search.RangeNodeBuilder
    effective_date: Search.RangeNodeBuilder
    id: Search.TextNodeBuilder
    kind: Search.MultipleValueNodeBuilder
    merchant_account_id: Search.MultipleValueNodeBuilder
    pre_dispute_program: Search.MultipleValueNodeBuilder
    reason: Search.MultipleValueNodeBuilder
    reason_code: Search.MultipleValueNodeBuilder
    received_date: Search.RangeNodeBuilder
    reference_number: Search.TextNodeBuilder
    reply_by_date: Search.RangeNodeBuilder
    status: Search.MultipleValueNodeBuilder
    transaction_id: Search.TextNodeBuilder
    transaction_source: Search.MultipleValueNodeBuilder



from icalendar.cal import Component

component = Component()
component.add("summary", "Test 1")
component.add("dtstart", "2022-01-01", encode=True)
component.add("dtend", "2022-01-02", encode=False)
component.add("location", "Test 3", {})
component.add("dtstamp", "2022-01-03", parameters={}, encode=True)
component.add("description", "Test 2", parameters={}, encode=False)  # type: ignore



from abc import ABCMeta, abstractmethod
from collections.abc import Mapping
from typing import ClassVar

default_max_pbkdf2_iterations: int

class JWAAlgorithm(metaclass=ABCMeta):
    @property
    @abstractmethod
    def name(self) -> str: ...
    @property
    @abstractmethod
    def description(self) -> str: ...
    @property
    @abstractmethod
    def keysize(self) -> int: ...
    @property
    @abstractmethod
    def algorithm_usage_location(self) -> str: ...
    @property
    @abstractmethod
    def algorithm_use(self) -> str: ...
    @property
    def input_keysize(self) -> int: ...

class JWA:
    algorithms_registry: ClassVar[Mapping[str, JWAAlgorithm]]
    @classmethod
    def instantiate_alg(cls, name: str, use: str | None = None) -> JWAAlgorithm: ...
    @classmethod
    def signing_alg(cls, name: str) -> JWAAlgorithm: ...
    @classmethod
    def keymgmt_alg(cls, name: str) -> JWAAlgorithm: ...
    @classmethod
    def encryption_alg(cls, name: str) -> JWAAlgorithm: ...



from collections.abc import Iterable
from typing import NoReturn, Protocol, type_check_only

from docker import APIClient
from docker.models.configs import ConfigCollection
from docker.models.containers import ContainerCollection
from docker.models.images import ImageCollection
from docker.models.networks import NetworkCollection
from docker.models.nodes import NodeCollection
from docker.models.plugins import PluginCollection
from docker.models.secrets import SecretCollection
from docker.models.services import ServiceCollection
from docker.models.swarm import Swarm
from docker.models.volumes import VolumeCollection

@type_check_only
class _Environ(Protocol):
    def __getitem__(self, k: str, /) -> str: ...
    def keys(self) -> Iterable[str]: ...

class DockerClient:
    api: APIClient
    def __init__(self, *args, **kwargs) -> None: ...
    @classmethod
    def from_env(
        cls,
        *,
        version: str | None = None,
        timeout: int = ...,
        max_pool_size: int = ...,
        environment: _Environ | None = None,
        use_ssh_client: bool = False,
    ) -> DockerClient: ...
    @property
    def configs(self) -> ConfigCollection: ...
    @property
    def containers(self) -> ContainerCollection: ...
    @property
    def images(self) -> ImageCollection: ...
    @property
    def networks(self) -> NetworkCollection: ...
    @property
    def nodes(self) -> NodeCollection: ...
    @property
    def plugins(self) -> PluginCollection: ...
    @property
    def secrets(self) -> SecretCollection: ...
    @property
    def services(self) -> ServiceCollection: ...
    @property
    def swarm(self) -> Swarm: ...
    @property
    def volumes(self) -> VolumeCollection: ...
    def events(self, *args, **kwargs): ...
    def df(self): ...
    def info(self, *args, **kwargs): ...
    def login(self, *args, **kwargs): ...
    def ping(self, *args, **kwargs): ...
    def version(self, *args, **kwargs): ...
    def close(self): ...
    def __getattr__(self, name: str) -> NoReturn: ...

from_env = DockerClient.from_env



from _typeshed import Incomplete

class AuthorizationServerMetadata(dict[str, object]):
    REGISTRY_KEYS: Incomplete
    def validate_issuer(self) -> None: ...
    def validate_authorization_endpoint(self) -> None: ...
    def validate_token_endpoint(self) -> None: ...
    def validate_jwks_uri(self) -> None: ...
    def validate_registration_endpoint(self) -> None: ...
    def validate_scopes_supported(self) -> None: ...
    def validate_response_types_supported(self) -> None: ...
    def validate_response_modes_supported(self) -> None: ...
    def validate_grant_types_supported(self) -> None: ...
    def validate_token_endpoint_auth_methods_supported(self) -> None: ...
    def validate_token_endpoint_auth_signing_alg_values_supported(self) -> None: ...
    def validate_service_documentation(self) -> None: ...
    def validate_ui_locales_supported(self) -> None: ...
    def validate_op_policy_uri(self) -> None: ...
    def validate_op_tos_uri(self) -> None: ...
    def validate_revocation_endpoint(self) -> None: ...
    def validate_revocation_endpoint_auth_methods_supported(self) -> None: ...
    def validate_revocation_endpoint_auth_signing_alg_values_supported(self) -> None: ...
    def validate_introspection_endpoint(self) -> None: ...
    def validate_introspection_endpoint_auth_methods_supported(self) -> None: ...
    def validate_introspection_endpoint_auth_signing_alg_values_supported(self) -> None: ...
    def validate_code_challenge_methods_supported(self) -> None: ...
    @property
    def response_modes_supported(self): ...
    @property
    def grant_types_supported(self): ...
    @property
    def token_endpoint_auth_methods_supported(self): ...
    @property
    def revocation_endpoint_auth_methods_supported(self): ...
    @property
    def introspection_endpoint_auth_methods_supported(self): ...
    def validate(self) -> None: ...
    def __getattr__(self, key): ...

def validate_array_value(metadata, key) -> None: ...



from _typeshed import Unused
from typing import TypedDict, type_check_only
from typing_extensions import NotRequired

@type_check_only
class _Thread(TypedDict):
    id: int
    name: NotRequired[str]

@type_check_only
class _ThreadsResult(TypedDict):
    threads: list[_Thread]

def threads(**args: Unused) -> _ThreadsResult: ...



from _typeshed import Incomplete
from typing_extensions import Self

from braintree.graphql.inputs.monetary_amount_input import MonetaryAmountInput
from braintree.graphql.inputs.paypal_payee_input import PayPalPayeeInput

class PayPalPurchaseUnitInput:
    def __init__(self, amount: MonetaryAmountInput | None = None, payee: PayPalPayeeInput | None = None) -> None: ...
    def to_graphql_variables(self) -> dict[str, Incomplete]: ...
    @staticmethod
    def builder(amount: MonetaryAmountInput) -> Builder: ...

    class Builder:
        def __init__(self, amount: MonetaryAmountInput) -> None: ...
        def payee(self, payee: PayPalPayeeInput) -> Self: ...
        def build(self) -> PayPalPurchaseUnitInput: ...



from _typeshed import Incomplete
from typing import NoReturn, overload
from typing_extensions import deprecated

from networkx.classes.graph import Graph, _Node
from networkx.classes.multigraph import MultiGraph
from networkx.utils.backends import _dispatchable

__all__ = ["CurrentEdge", "Level", "GlobalRelabelThreshold", "build_residual_network", "detect_unboundedness", "build_flow_dict"]

class CurrentEdge:
    __slots__ = ("_edges", "_it", "_curr")
    def __init__(self, edges) -> None: ...
    def get(self): ...
    def move_to_next(self) -> None: ...

class Level:
    __slots__ = ("active", "inactive")
    active: Incomplete
    inactive: Incomplete

    def __init__(self) -> None: ...

class GlobalRelabelThreshold:
    def __init__(self, n, m, freq) -> None: ...
    def add_work(self, work) -> None: ...
    def is_reached(self) -> bool: ...
    def clear_work(self) -> None: ...

@overload
@deprecated("MultiGraph and MultiDiGraph not supported (yet).")
def build_residual_network(G: MultiGraph[_Node], capacity, *, backend: str | None = None, **backend_kwargs) -> NoReturn: ...
@overload
def build_residual_network(G: Graph[_Node], capacity, *, backend: str | None = None, **backend_kwargs): ...
@_dispatchable
def detect_unboundedness(R, s, t) -> None: ...
@_dispatchable
def build_flow_dict(G: Graph[_Node], R): ...



from _typeshed import StrOrBytesPath
from typing import Any, Final

LIST_OF_FILE_NAMES: Final[list[str]]

def get_extension(srcfilename, modname, sources=(), **kwds): ...
def compile(tmpdir, ext, compiler_verbose: int = 0, debug=None): ...
def maybe_relative_path(path: StrOrBytesPath) -> StrOrBytesPath | str: ...

int_or_long = int

def flatten(x: int | str | list[Any] | tuple[Any] | dict[Any, Any]) -> str: ...



from collections.abc import Callable, Hashable
from typing import Literal
from typing_extensions import TypeAlias

from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

_Algorithm: TypeAlias = Literal["greedy++", "fista"]

__all__ = ["densest_subgraph"]

ALGORITHMS: dict[_Algorithm, Callable[[Graph[Hashable], int], tuple[float, set[int]]]]

@_dispatchable
def densest_subgraph(G: Graph[_Node], iterations: int = 1, *, method: _Algorithm = "fista") -> tuple[float, set[int]]: ...



from collections.abc import Iterator, Sequence
from typing import Any, overload
from typing_extensions import disjoint_base

@disjoint_base
class ResultRow:
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    column_names: tuple[str, ...]
    column_values: tuple[Any, ...]

    def __len__(self) -> int: ...
    @overload
    def __getitem__(self, index: int, /) -> Any: ...
    @overload
    def __getitem__(self, index: slice, /) -> Sequence[Any]: ...
    def __iter__(self) -> Iterator[Any]: ...
    # __next__, __delitem__, __setitem__ are technically defined but lead always to an error



import re
from collections.abc import Collection
from typing import ClassVar, NamedTuple
from xml.etree.ElementTree import Element

from markdown import util
from markdown.core import Markdown

def build_inlinepatterns(md: Markdown, **kwargs) -> util.Registry[Pattern]: ...

NOIMG: str
BACKTICK_RE: str
ESCAPE_RE: str
EMPHASIS_RE: str
STRONG_RE: str
SMART_STRONG_RE: str
SMART_EMPHASIS_RE: str
SMART_STRONG_EM_RE: str
EM_STRONG_RE: str
EM_STRONG2_RE: str
STRONG_EM_RE: str
STRONG_EM2_RE: str
STRONG_EM3_RE: str
LINK_RE: str
IMAGE_LINK_RE: str
REFERENCE_RE: str
IMAGE_REFERENCE_RE: str
NOT_STRONG_RE: str
AUTOLINK_RE: str
AUTOMAIL_RE: str
HTML_RE: str
ENTITY_RE: str
LINE_BREAK_RE: str

def dequote(string: str) -> str: ...

class EmStrongItem(NamedTuple):
    pattern: re.Pattern[str]
    builder: str
    tags: str

class Pattern:
    ANCESTOR_EXCLUDES: ClassVar[Collection[str]]
    pattern: str
    compiled_re: re.Pattern[str]
    md: Markdown
    def __init__(self, pattern: str, md: Markdown | None = None) -> None: ...
    def getCompiledRegExp(self) -> re.Pattern[str]: ...
    def handleMatch(self, m: re.Match[str]) -> str | Element | None: ...
    def type(self) -> str: ...
    def unescape(self, text: str) -> str: ...

class InlineProcessor(Pattern):
    safe_mode: bool
    def __init__(self, pattern: str, md: Markdown | None = None) -> None: ...
    def handleMatch(self, m: re.Match[str], data: str) -> tuple[Element | str | None, int | None, int | None]: ...  # type: ignore[override]

class SimpleTextPattern(Pattern): ...
class SimpleTextInlineProcessor(InlineProcessor): ...
class EscapeInlineProcessor(InlineProcessor): ...

class SimpleTagPattern(Pattern):
    tag: str
    def __init__(self, pattern: str, tag: str) -> None: ...

class SimpleTagInlineProcessor(InlineProcessor):
    tag: str
    def __init__(self, pattern: str, tag: str) -> None: ...

class SubstituteTagPattern(SimpleTagPattern): ...
class SubstituteTagInlineProcessor(SimpleTagInlineProcessor): ...

class BacktickInlineProcessor(InlineProcessor):
    ESCAPED_BSLASH: str
    tag: str
    def __init__(self, pattern: str) -> None: ...

class DoubleTagPattern(SimpleTagPattern): ...
class DoubleTagInlineProcessor(SimpleTagInlineProcessor): ...

class HtmlInlineProcessor(InlineProcessor):
    def unescape(self, text: str) -> str: ...
    def backslash_unescape(self, text: str) -> str: ...

class AsteriskProcessor(InlineProcessor):
    PATTERNS: ClassVar[list[EmStrongItem]]
    def build_single(self, m: re.Match[str], tag: str, idx: int) -> Element: ...
    def build_double(self, m: re.Match[str], tags: str, idx: int) -> Element: ...
    def build_double2(self, m: re.Match[str], tags: str, idx: int) -> Element: ...
    def parse_sub_patterns(self, data: str, parent: Element, last: Element | None, idx: int) -> None: ...
    def build_element(self, m: re.Match[str], builder: str, tags: str, index: int) -> Element: ...

class UnderscoreProcessor(AsteriskProcessor): ...

class LinkInlineProcessor(InlineProcessor):
    RE_LINK: ClassVar[re.Pattern[str]]
    RE_TITLE_CLEAN: ClassVar[re.Pattern[str]]
    def getLink(self, data: str, index: int) -> tuple[str, str | None, int, bool]: ...
    def getText(self, data: str, index: int) -> tuple[str, int, bool]: ...

class ImageInlineProcessor(LinkInlineProcessor): ...

class ReferenceInlineProcessor(LinkInlineProcessor):
    NEWLINE_CLEANUP_RE: ClassVar[re.Pattern[str]]
    def evalId(self, data: str, index: int, text: str) -> tuple[str | None, int, bool]: ...
    def makeTag(self, href: str, title: str, text: str) -> Element: ...

class ShortReferenceInlineProcessor(ReferenceInlineProcessor): ...
class ImageReferenceInlineProcessor(ReferenceInlineProcessor): ...
class ShortImageReferenceInlineProcessor(ImageReferenceInlineProcessor): ...
class AutolinkInlineProcessor(InlineProcessor): ...
class AutomailInlineProcessor(InlineProcessor): ...



from _typeshed import Incomplete

from .base import AuthenticationBase

class BackChannelLogin(AuthenticationBase):
    def back_channel_login(
        self,
        binding_message: str,
        login_hint: str,
        scope: str,
        authorization_details: str | list[dict[str, Incomplete]] | None = None,
        **kwargs,
    ): ...



import itertools

flatten = itertools.chain.from_iterable

class Installer:
    nspkg_ext: str
    def install_namespaces(self) -> None: ...
    def uninstall_namespaces(self) -> None: ...

class DevelopInstaller(Installer): ...



from collections.abc import Iterable
from typing import SupportsIndex, TypeVar
from typing_extensions import Self, TypeAlias

_T = TypeVar("_T")

class BarrelList(list[_T]):
    lists: list[list[_T]]
    def __init__(self, iterable: Iterable[_T] | None = None) -> None: ...
    def insert(self, index: SupportsIndex, item: _T) -> None: ...
    def append(self, item: _T) -> None: ...
    def extend(self, iterable: Iterable[_T]) -> None: ...
    def pop(self, *a) -> _T: ...
    def iter_slice(self, start: int | None, stop: int | None, step: int | None = None) -> Iterable[_T]: ...
    def del_slice(self, start: int, stop: int, step: int | None = None) -> None: ...
    __delslice__ = del_slice
    @classmethod
    def from_iterable(cls, it: Iterable[_T]) -> Self: ...
    def __getslice__(self, start: int, stop: int): ...
    def __setslice__(self, start: SupportsIndex, stop: SupportsIndex, sequence: Iterable[_T]) -> None: ...
    def sort(self) -> None: ...  # type: ignore[override]
    def reverse(self) -> None: ...
    def count(self, item: _T) -> int: ...
    def index(self, item: _T) -> int: ...  # type: ignore[override]

BList: TypeAlias = BarrelList[_T]

class SplayList(list[_T]):
    def shift(self, item_index: int, dest_index: int = 0) -> None: ...
    def swap(self, item_index: SupportsIndex, dest_index: SupportsIndex) -> None: ...

__all__ = ["BList", "BarrelList"]



from _typeshed import Incomplete, Unused
from typing import ClassVar, Literal
from typing_extensions import TypeAlias

from openpyxl.descriptors.base import Alias, NoneSet, Typed, _ConvertibleToBool
from openpyxl.descriptors.nested import EmptyTag
from openpyxl.descriptors.serialisable import Serialisable
from openpyxl.drawing.colors import ColorChoice, ColorChoiceDescriptor
from openpyxl.drawing.fill import GradientFillProperties, PatternFillProperties
from openpyxl.drawing.geometry import CustomGeometry2D, PresetGeometry2D, Scene3D, Shape3D, Transform2D
from openpyxl.drawing.line import LineProperties

from ..xml._functions_overloads import _HasTagAndGet

_GraphicalPropertiesBwMode: TypeAlias = Literal[
    "clr", "auto", "gray", "ltGray", "invGray", "grayWhite", "blackGray", "blackWhite", "black", "white", "hidden"
]

class GraphicalProperties(Serialisable):
    tagname: ClassVar[str]
    bwMode: NoneSet[_GraphicalPropertiesBwMode]
    xfrm: Typed[Transform2D, Literal[True]]
    transform: Alias
    custGeom: Typed[CustomGeometry2D, Literal[True]]
    prstGeom: Typed[PresetGeometry2D, Literal[True]]
    noFill: EmptyTag[Literal[False]]
    solidFill: ColorChoiceDescriptor
    gradFill: Typed[GradientFillProperties, Literal[True]]
    pattFill: Typed[PatternFillProperties, Literal[True]]
    ln: Typed[LineProperties, Literal[True]]
    line: Alias
    scene3d: Typed[Scene3D, Literal[True]]
    sp3d: Typed[Shape3D, Literal[True]]
    shape3D: Alias
    extLst: Typed[Incomplete, Literal[True]]
    __elements__: ClassVar[tuple[str, ...]]
    def __init__(
        self,
        bwMode: _GraphicalPropertiesBwMode | Literal["none"] | None = None,
        xfrm: Transform2D | None = None,
        noFill: _HasTagAndGet[_ConvertibleToBool] | _ConvertibleToBool = None,
        solidFill: str | ColorChoice | None = None,
        gradFill: GradientFillProperties | None = None,
        pattFill: PatternFillProperties | None = None,
        ln=None,
        scene3d: Scene3D | None = None,
        custGeom: CustomGeometry2D | None = None,
        prstGeom: PresetGeometry2D | None = None,
        sp3d: Shape3D | None = None,
        extLst: Unused = None,
    ) -> None: ...



from _typeshed import Incomplete
from collections.abc import Generator

from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = ["read_pajek", "parse_pajek", "generate_pajek", "write_pajek"]

def generate_pajek(G: Graph[_Node]) -> Generator[Incomplete, None, None]: ...
def write_pajek(G: Graph[_Node], path, encoding: str = "UTF-8") -> None: ...
@_dispatchable
def read_pajek(path, encoding: str = "UTF-8"): ...
@_dispatchable
def parse_pajek(lines): ...
def make_qstr(t): ...



"""Utility functions and classes for Jac compilation toolchain."""

import dis
import inspect
import marshal
import os
import pdb
import re
from collections.abc import Callable, Sequence
from dataclasses import fields, is_dataclass
from functools import lru_cache
from traceback import TracebackException
from typing import TYPE_CHECKING, get_args, get_origin

if TYPE_CHECKING:
    import pluggy


@lru_cache(maxsize=256)
def pascal_to_snake(pascal_string: str) -> str:
    """Convert pascal case to snake case."""
    snake_string = re.sub(r"(?<!^)(?=[A-Z])", "_", pascal_string).lower()
    return snake_string


def heading_to_snake(heading: str) -> str:
    """Convert string to snakecase including replacing(/ ,- )."""
    return heading.strip().replace("-", "_").replace("/", "_").replace(" ", "_").lower()


def add_line_numbers(s: str) -> str:
    """Add line numbers to a string."""
    lines = s.split("\n")
    return "\n".join(f"{i + 1}: \t{line}" for i, line in enumerate(lines))


def get_uni_nodes_as_snake_case() -> list[str]:
    """Get all AST nodes as snake case."""
    import inspect
    import sys

    import jaclang.pycore.unitree as uni

    module_name = uni.__name__
    module = sys.modules[module_name]

    # Retrieve the source code of the module
    source_code = inspect.getsource(module)

    classes = inspect.getmembers(module, inspect.isclass)
    uni_node_classes = [cls for _, cls in classes if issubclass(cls, uni.UniNode)]

    ordered_classes = sorted(
        uni_node_classes, key=lambda cls: source_code.find(f"class {cls.__name__}")
    )
    snake_names = []
    for cls in ordered_classes:
        class_name = cls.__name__
        snake_names.append(pascal_to_snake(class_name))
    return snake_names


def dump_traceback(e: Exception) -> str:
    """Dump the stack frames of the exception."""
    trace_dump = ""

    # Utility function to get the error line char offset.
    def byte_offset_to_char_offset(string: str, offset: int) -> int:
        return len(string.encode("utf-8")[:offset].decode("utf-8", errors="replace"))

    # Utility function to check if a file is a compiled Jac file and get the original .jac source
    def get_jac_source_info(py_filename: str) -> tuple[str | None, str | None]:
        """Return (jac_filename, jac_source) if available, else (None, None)."""
        # Currently no mapping from Python to Jac source files is available
        return None, None

    tb = TracebackException(type(e), e, e.__traceback__, limit=None, compact=True)
    trace_dump += f"Error: {str(e)}\n"

    # The first frame is the call the to the above `exec` function, not usefull to the enduser,
    # and Make the most recent call first.
    tb.stack.pop(0)
    tb.stack.reverse()

    # FIXME: should be some settings, we should replace to ensure the anchors length match.
    dump_tab_width = 2

    # Helper function to check if a frame is part of the internal runtime
    def is_internal_runtime_frame(filename: str) -> bool:
        """Check if a frame is part of the Jac internal runtime."""
        normalized: str = filename.replace("\\", "/")
        return (
            "/jaclang/runtimelib/" in normalized
            or "/jaclang/pycore/" in normalized
            or "/jaclang/vendor/" in normalized
            or "/site-packages/pluggy/" in normalized
            or normalized.endswith("/jaclang/meta_importer.py")
        )

    # Process and print frames, collapsing consecutive internal runtime calls
    seen_runtime_marker: bool = False
    collapse_internal: bool = True

    for idx, frame in enumerate(tb.stack):
        is_internal: bool = is_internal_runtime_frame(frame.filename)

        # Collapse consecutive internal runtime frames into a single marker (if enabled)
        if is_internal and collapse_internal:
            if not seen_runtime_marker:
                trace_dump += f"\n{' ' * dump_tab_width}... [internal runtime calls]"
                seen_runtime_marker = True
            continue

        # Reset the marker flag when we hit a user frame
        seen_runtime_marker = False

        func_signature: str = frame.name + ("()" if frame.name.isidentifier() else "")

        # Check if we can map this to a .jac file
        jac_filename, jac_source = get_jac_source_info(frame.filename)
        display_filename: str = jac_filename if jac_filename else frame.filename
        display_source: str | None = jac_source if jac_source else None

        # Pretty print the most recent call's location.
        if idx == 0 and (
            (frame.lineno is not None) and frame.line and frame.line.strip() != ""
        ):
            # Note: This is CPython internals we're trying to get since python doesn't provide
            # the frames original line but the stripped version so we had to do this.
            line_o = frame.line  # Fallback line.
            if hasattr(frame, "_original_line"):
                line_o = frame._original_line.rstrip()  # type: ignore [attr-defined]
            elif hasattr(frame, "_original_lines"):
                # https://github.com/python/cpython/issues/106922
                line_o = frame._original_lines.split("\n")[0].rstrip()  # type: ignore [attr-defined]

            if frame.colno is not None and frame.end_colno is not None:
                off_start = byte_offset_to_char_offset(line_o, frame.colno) - 1
                off_end = byte_offset_to_char_offset(line_o, frame.end_colno) - 1

                # Get the source - prefer .jac source if available, otherwise use .py
                file_source = display_source
                if file_source is None:
                    try:
                        with open(frame.filename) as file:
                            file_source = file.read()
                    except Exception:
                        file_source = ""

                if file_source:
                    # Get the source offset.
                    lines = file_source.split("\n")
                    for i in range(frame.lineno - 1):
                        off_start += len(lines[i]) + 1
                        off_end += len(lines[i]) + 1

                    trace_dump += pretty_print_source_location(
                        display_filename, file_source, frame.lineno, off_start, off_end
                    )

        trace_dump += f"\n{' ' * dump_tab_width}at {func_signature} {display_filename}:{frame.lineno}"

    return trace_dump


# FIXME: Use a proper color library and/or move this somewhere common to jac stack and use it everywhere.
# Reference: https://gist.github.com/rene-d/9e584a7dd2935d0f461904b9f2950007
class ANSIColors:
    """ANSI color codes."""

    BLACK = "\033[0;30m"
    RED = "\033[0;31m"
    GREEN = "\033[0;32m"
    BROWN = "\033[0;33m"
    BLUE = "\033[0;34m"
    PURPLE = "\033[0;35m"
    CYAN = "\033[0;36m"
    LIGHT_GRAY = "\033[0;37m"
    DARK_GRAY = "\033[1;30m"
    LIGHT_RED = "\033[1;31m"
    LIGHT_GREEN = "\033[1;32m"
    YELLOW = "\033[1;33m"
    LIGHT_BLUE = "\033[1;34m"
    LIGHT_PURPLE = "\033[1;35m"
    LIGHT_CYAN = "\033[1;36m"
    LIGHT_WHITE = "\033[1;37m"
    BOLD = "\033[1m"
    FAINT = "\033[2m"
    ITALIC = "\033[3m"
    UNDERLINE = "\033[4m"
    BLINK = "\033[5m"
    NEGATIVE = "\033[7m"
    CROSSED = "\033[9m"
    END = "\033[0m"


# TODO: After implementing the TextRange (or simillar named) class to mark a text range
# refactor the parameter to accept an instace of that text range object.
def pretty_print_source_location(
    file_path: str,
    file_source: str,
    error_line: int,
    pos_start: int,
    pos_end: int,
    *,
    colors: bool = False,
) -> str:
    """Pretty print internal method for the pretty_print method."""
    # NOTE: The Line numbers and the column numbers are starts with 1.
    # We print totally 5 lines (error line and above 2 and bellow 2).

    # The width of the line number we'll be printing (more of a settings).
    line_num_width: int = 5

    idx: int = pos_start  # Pointer for the current character.

    if file_source == "" or file_path == "":
        return ""

    start_line: int = error_line - 2
    if start_line < 1:
        start_line = 1
    end_line: int = start_line + 5  # Index is exclusive.

    # Get the first character of the [start_line].
    file_source.splitlines(True)[start_line - 1]
    curr_line: int = error_line
    while idx >= 0 and curr_line >= start_line:
        idx -= 1
        if idx < 0:
            break
        if file_source[idx] == "\n":
            curr_line -= 1

    idx += 1  # Enter the line.
    assert idx == 0 or file_source[idx - 1] == "\n"

    pretty_dump = ""

    # Print each lines.
    curr_line = start_line
    while curr_line < end_line:
        pretty_dump += f"%{line_num_width}d | " % curr_line

        idx_line_start = idx
        while idx < len(file_source) and file_source[idx] != "\n":
            idx += 1  # Run to the line end.

        if colors and (curr_line == error_line):
            pretty_dump += (
                file_source[idx_line_start:pos_start]
                + f"{ANSIColors.RED}{file_source[pos_start:pos_end]}{ANSIColors.END}"
                + file_source[pos_end:idx]
            )
        else:
            pretty_dump += file_source[idx_line_start:idx]

        pretty_dump += "\n"

        if curr_line == error_line:  # Print the current line with indicator.
            pretty_dump += f"%{line_num_width}s | " % " "

            spaces = ""
            for idx_pre in range(idx_line_start, pos_start):
                spaces += "\t" if file_source[idx_pre] == "\t" else " "

            err_token_len = pos_end - pos_start
            underline = "^" * err_token_len
            if colors:
                underline = f"{ANSIColors.RED}{underline}{ANSIColors.END}"
            pretty_dump += spaces + underline + "\n"

        if idx == len(file_source):
            break
        curr_line += 1
        idx += 1

    return pretty_dump[:-1]  # Get rid of the last newline (of the last line).


class Jdb(pdb.Pdb):
    """Jac debugger."""

    def __init__(self, *args, **kwargs) -> None:  # noqa
        """Initialize the Jac debugger."""
        super().__init__(*args, **kwargs)
        self.prompt = "Jdb > "

    def has_breakpoint(self, bytecode: bytes) -> bool:
        """Check for breakpoint."""
        code = marshal.loads(bytecode)
        instructions = dis.get_instructions(code)
        return any(
            instruction.opname in ("LOAD_GLOBAL", "LOAD_NAME", "LOAD_FAST")
            and instruction.argval == "breakpoint"
            for instruction in instructions
        )


debugger = Jdb()


def read_file_with_encoding(file_path: str) -> str:
    """Read file with proper encoding detection.

    Tries multiple encodings to handle files with different encodings.
    """
    encodings_to_try = [
        "utf-8-sig",
        "utf-8",
        "utf-16",
        "utf-16le",
        "utf-16be",
    ]

    for encoding in encodings_to_try:
        try:
            with open(file_path, encoding=encoding) as f:
                return f.read()
        except UnicodeError:
            continue
        except Exception as e:
            raise OSError(
                f"Could not read file {file_path}: {e}. "
                f"Report this issue: https://github.com/jaseci-labs/jaseci/issues"
            ) from e

    raise OSError(
        f"Could not read file {file_path} with any encoding. "
        f"Report this issue: https://github.com/jaseci-labs/jaseci/issues"
    )


def _short_type_name(t: object) -> str:
    """Return a short, readable type name for annotations."""
    if t is inspect._empty:
        return "Any"

    origin = get_origin(t)
    if origin is not None:
        name = getattr(origin, "__name__", str(origin).replace("typing.", ""))
        args = get_args(t)
        if not args:
            return name
        return f"{name}[{','.join(_short_type_name(a) for a in args)}]"

    return getattr(t, "__name__", str(t).replace("typing.", ""))


def _signature_summary(func: Callable) -> str:
    """Summarize function signature as (T1,T2)->R (drop param names/self)."""
    try:
        sig = inspect.signature(func)
    except (TypeError, ValueError):
        return ""

    params: list[str] = []
    for p in sig.parameters.values():
        if p.name == "self":
            continue
        params.append(_short_type_name(p.annotation))

    ret = _short_type_name(sig.return_annotation)
    return f"({','.join(params)})->{ret}"


def _safe_repr(v: object, limit: int = 120) -> str:
    """Keep repr readable; avoids huge dumps while staying generic."""
    s = repr(v)
    return s if len(s) <= limit else s[: limit - 1] + "…"


def _describe_node(obj: object) -> str:
    """Single-line description used for LLM routing."""
    cls = obj.__class__

    # Attributes
    attrs: list[str] = []
    if is_dataclass(obj):
        for f in fields(obj):
            attrs.append(f"{f.name}={_safe_repr(getattr(obj, f.name))}")
    else:
        for k, v in vars(obj).items():
            if k.startswith("_"):
                continue
            attrs.append(f"{k}={_safe_repr(v)}")

    # Methods
    methods: list[str] = []
    for m_name, func in inspect.getmembers(cls, predicate=inspect.isfunction):
        if m_name.startswith("__"):
            continue
        sig = _signature_summary(func)
        methods.append(f"{m_name}{sig}" if sig else m_name)

    parts = [cls.__name__]
    if attrs:
        parts.append("attrs:" + ",".join(attrs))
    if methods:
        parts.append("methods:" + ",".join(methods))
    return " — ".join(parts)


def _describe_nodes_list(objects: Sequence[object]) -> str:
    """One object per line, index-friendly for returning list[int]."""
    return "\n".join(f"{i}) {_describe_node(obj)}" for i, obj in enumerate(objects))


# =============================================================================
# Plugin Loading Helpers
# =============================================================================


class DistFacade:
    """Minimal distribution facade for tracking plugin origins.

    Used when manually loading entry points to maintain compatibility
    with pluggy's list_plugin_distinfo() interface.
    """

    def __init__(self, dist: "object | None") -> None:
        self._dist = dist

    @property
    def project_name(self) -> str:
        """Get the distribution/package name."""
        return (
            self._dist.name if self._dist and hasattr(self._dist, "name") else "unknown"
        )

    @property
    def version(self) -> str:
        """Get the distribution version."""
        return (
            self._dist.version
            if self._dist and hasattr(self._dist, "version")
            else "0.0.0"
        )


def get_disabled_plugins() -> list[str]:
    """Get list of disabled plugins from JAC_DISABLED_PLUGINS env var or jac.toml.

    Priority: JAC_DISABLED_PLUGINS env var > jac.toml
    Supports qualified names (package:plugin), short names, and "*" wildcard.

    Returns:
        List of disabled plugin identifiers.
    """
    from pathlib import Path

    # First check environment variable (takes precedence)
    env_disabled = os.environ.get("JAC_DISABLED_PLUGINS", "").strip()
    if env_disabled:
        # Support comma-separated list or single value
        return [d.strip() for d in env_disabled.split(",") if d.strip()]

    # Fall back to jac.toml
    try:
        import tomllib
    except ImportError:
        # Python < 3.11 fallback
        try:
            import tomli as tomllib  # type: ignore
        except ImportError:
            return []

    # Search for jac.toml starting from current directory
    current = Path.cwd().resolve()
    while current != current.parent:
        toml_path = current / "jac.toml"
        if toml_path.exists():
            try:
                with open(toml_path, "rb") as f:
                    data = tomllib.load(f)
                return data.get("plugins", {}).get("disabled", [])
            except Exception:
                return []
        current = current.parent
    # Check root
    toml_path = current / "jac.toml"
    if toml_path.exists():
        try:
            with open(toml_path, "rb") as f:
                data = tomllib.load(f)
            return data.get("plugins", {}).get("disabled", [])
        except Exception:
            pass
    return []


def load_plugins_with_disabling(
    plugin_manager: "pluggy.PluginManager", disabled_list: list[str]
) -> None:
    """Load setuptools entry points while respecting disabled plugin list.

    Supports:
    - "*" wildcard to disable all external plugins
    - Qualified names (package:plugin) for specific plugin disabling
    - Package wildcards (package:*) to disable all plugins from a package
    - Short names for backwards compatibility (affects all packages with that name)

    This manually loads entry points to allow fine-grained control over which
    plugins are loaded when multiple packages have plugins with the same name.

    Args:
        plugin_manager: The pluggy PluginManager instance.
        disabled_list: List of plugin identifiers to disable.
    """
    from importlib.metadata import entry_points

    # Check for wildcard - disable all external plugins
    disable_all = "*" in disabled_list

    # Parse disabled list into qualified and short name sets
    disabled_qualified: set[str] = set()  # package:plugin format
    disabled_packages: set[str] = set()  # package:* format (all plugins from package)
    disabled_short: set[str] = set()  # legacy short names

    for name in disabled_list:
        if name == "*":
            continue  # Already handled above
        elif name.endswith(":*"):
            # Package wildcard: disable all plugins from this package
            disabled_packages.add(name[:-2])  # Remove ":*"
        elif ":" in name:
            disabled_qualified.add(name)
        else:
            disabled_short.add(name)

    # If disable_all, don't load any external plugins
    if disable_all:
        return

    # Get entry points for the "jac" group
    try:
        # Python 3.10+
        eps = entry_points(group="jac")
    except TypeError:
        # Python 3.9 fallback - entry_points() returns dict-like SelectableGroups
        all_eps = entry_points()
        eps = all_eps.get("jac", [])  # type: ignore[attr-defined]

    # Manually load each entry point, skipping disabled ones
    for ep in eps:
        # Get the distribution (package) this entry point comes from
        try:
            dist = ep.dist
            dist_name = dist.name if dist else "unknown"
        except Exception:
            dist = None
            dist_name = "unknown"

        qualified_name = f"{dist_name}:{ep.name}"

        # Check if this specific plugin should be disabled
        should_disable = False
        if dist_name in disabled_packages:
            # Package wildcard match
            should_disable = True
        elif qualified_name in disabled_qualified:
            should_disable = True
        elif ep.name in disabled_short:
            # Legacy: short name matches (affects all packages with this name)
            should_disable = True

        if should_disable:
            # Skip this plugin entirely
            continue

        # Load and register the plugin
        try:
            plugin = ep.load()
            # Check if already registered (another package with same name)
            if plugin_manager.is_registered(plugin):
                continue
            plugin_manager.register(plugin, name=ep.name)
            # Track distribution info for list_plugin_distinfo() to work
            plugin_manager._plugin_distinfo.append((plugin, DistFacade(dist)))
        except Exception:
            # Skip plugins that fail to load
            pass



import xml.etree.ElementTree as ET
from typing import ClassVar, Final

from docutils import nodes, parsers

__docformat__: Final = "reStructuredText"

class Parser(parsers.Parser):
    config_section_dependencies: ClassVar[tuple[str, ...]]
    settings_default_overrides: ClassVar[dict[str, bool]]

class Unknown(nodes.Special, nodes.Inline, nodes.Element): ...

def parse_element(inputstring: str, document: nodes.document | None = None) -> nodes.Element: ...
def element2node(element: ET.Element | None, document: nodes.document | None = None, unindent: bool = True) -> nodes.Element: ...
def append_text(node: nodes.Element, text: str | None, unindent: bool | None) -> None: ...



from _typeshed import ConvertibleToInt, Incomplete, Unused
from typing import ClassVar, Literal
from typing_extensions import TypeAlias

from openpyxl.chart.shapes import GraphicalProperties
from openpyxl.chart.text import RichText
from openpyxl.descriptors.base import Alias, Typed, _ConvertibleToBool
from openpyxl.descriptors.excel import ExtensionList
from openpyxl.descriptors.nested import NestedBool, NestedInteger, NestedNoneSet, NestedString, _NestedNoneSetParam
from openpyxl.descriptors.serialisable import Serialisable

from ..xml._functions_overloads import _HasTagAndGet

_DataLabelBaseDLblPos: TypeAlias = Literal["bestFit", "b", "ctr", "inBase", "inEnd", "l", "outEnd", "r", "t"]

class _DataLabelBase(Serialisable):
    numFmt: NestedString[Literal[True]]
    spPr: Typed[GraphicalProperties, Literal[True]]
    graphicalProperties: Alias
    txPr: Typed[RichText, Literal[True]]
    textProperties: Alias
    dLblPos: NestedNoneSet[_DataLabelBaseDLblPos]
    position: Alias
    showLegendKey: NestedBool[Literal[True]]
    showVal: NestedBool[Literal[True]]
    showCatName: NestedBool[Literal[True]]
    showSerName: NestedBool[Literal[True]]
    showPercent: NestedBool[Literal[True]]
    showBubbleSize: NestedBool[Literal[True]]
    showLeaderLines: NestedBool[Literal[True]]
    separator: NestedString[Literal[True]]
    extLst: Typed[ExtensionList, Literal[True]]
    __elements__: ClassVar[tuple[str, ...]]
    def __init__(
        self,
        numFmt: object = None,
        spPr: GraphicalProperties | None = None,
        txPr: RichText | None = None,
        dLblPos: _NestedNoneSetParam[_DataLabelBaseDLblPos] = None,
        showLegendKey: _HasTagAndGet[_ConvertibleToBool | None] | _ConvertibleToBool | None = None,
        showVal: _HasTagAndGet[_ConvertibleToBool | None] | _ConvertibleToBool | None = None,
        showCatName: _HasTagAndGet[_ConvertibleToBool | None] | _ConvertibleToBool | None = None,
        showSerName: _HasTagAndGet[_ConvertibleToBool | None] | _ConvertibleToBool | None = None,
        showPercent: _HasTagAndGet[_ConvertibleToBool | None] | _ConvertibleToBool | None = None,
        showBubbleSize: _HasTagAndGet[_ConvertibleToBool | None] | _ConvertibleToBool | None = None,
        showLeaderLines: _HasTagAndGet[_ConvertibleToBool | None] | _ConvertibleToBool | None = None,
        separator: object = None,
        extLst: Unused = None,
    ) -> None: ...

class DataLabel(_DataLabelBase):
    tagname: ClassVar[str]
    idx: NestedInteger[Literal[False]]
    # Same as parent
    # numFmt = _DataLabelBase.numFmt
    # spPr = _DataLabelBase.spPr
    # txPr = _DataLabelBase.txPr
    # dLblPos = _DataLabelBase.dLblPos
    # showLegendKey = _DataLabelBase.showLegendKey
    # showVal = _DataLabelBase.showVal
    # showCatName = _DataLabelBase.showCatName
    # showSerName = _DataLabelBase.showSerName
    # showPercent = _DataLabelBase.showPercent
    # showBubbleSize = _DataLabelBase.showBubbleSize
    # showLeaderLines = _DataLabelBase.showLeaderLines
    # separator = _DataLabelBase.separator
    # extLst = _DataLabelBase.extLst
    __elements__: ClassVar[tuple[str, ...]]
    def __init__(self, idx: _HasTagAndGet[ConvertibleToInt] | ConvertibleToInt = 0, **kw) -> None: ...

class DataLabelList(_DataLabelBase):
    tagname: ClassVar[str]
    dLbl: Incomplete
    delete: NestedBool[Literal[True]]
    # Same as parent
    # numFmt = _DataLabelBase.numFmt
    # spPr = _DataLabelBase.spPr
    # txPr = _DataLabelBase.txPr
    # dLblPos = _DataLabelBase.dLblPos
    # showLegendKey = _DataLabelBase.showLegendKey
    # showVal = _DataLabelBase.showVal
    # showCatName = _DataLabelBase.showCatName
    # showSerName = _DataLabelBase.showSerName
    # showPercent = _DataLabelBase.showPercent
    # showBubbleSize = _DataLabelBase.showBubbleSize
    # showLeaderLines = _DataLabelBase.showLeaderLines
    # separator = _DataLabelBase.separator
    # extLst = _DataLabelBase.extLst
    __elements__: ClassVar[tuple[str, ...]]
    def __init__(
        self, dLbl=(), delete: _HasTagAndGet[_ConvertibleToBool | None] | _ConvertibleToBool | None = None, **kw
    ) -> None: ...



from __future__ import annotations

import dataclasses as dc
import sys
from typing import TYPE_CHECKING, Any, Dict, FrozenSet, Tuple, Type, Union
from typing_extensions import Annotated, assert_type

if TYPE_CHECKING:
    from _typeshed import DataclassInstance


@dc.dataclass
class Foo:
    attr: str


assert_type(dc.fields(Foo), Tuple[dc.Field[Any], ...])

dc.asdict(Foo)  # type: ignore
dc.astuple(Foo)  # type: ignore
dc.replace(Foo)  # type: ignore

# See #9723 for why we can't make this assertion
# if dc.is_dataclass(Foo):
#     assert_type(Foo, Type[Foo])

f = Foo(attr="attr")

assert_type(dc.fields(f), Tuple[dc.Field[Any], ...])
assert_type(dc.asdict(f), Dict[str, Any])
assert_type(dc.astuple(f), Tuple[Any, ...])
assert_type(dc.replace(f, attr="new"), Foo)

if dc.is_dataclass(f):
    # The inferred type doesn't change
    # if it's already known to be a subtype of _DataclassInstance
    assert_type(f, Foo)


def is_dataclass_any(arg: Any) -> None:
    if dc.is_dataclass(arg):
        assert_type(arg, Union["DataclassInstance", Type["DataclassInstance"]])


def is_dataclass_object(arg: object) -> None:
    if dc.is_dataclass(arg):
        assert_type(arg, Union["DataclassInstance", Type["DataclassInstance"]])


def is_dataclass_type(arg: type) -> None:
    if dc.is_dataclass(arg):
        assert_type(arg, Type["DataclassInstance"])


def check_other_isdataclass_overloads(x: type, y: object) -> None:
    # TODO: neither pyright nor mypy emit error on this -- why?
    # dc.fields(x)

    dc.fields(y)  # type: ignore

    dc.asdict(x)  # type: ignore
    dc.asdict(y)  # type: ignore

    dc.astuple(x)  # type: ignore
    dc.astuple(y)  # type: ignore

    dc.replace(x)  # type: ignore
    dc.replace(y)  # type: ignore

    if dc.is_dataclass(x):
        assert_type(x, Type["DataclassInstance"])
        assert_type(dc.fields(x), Tuple[dc.Field[Any], ...])

        dc.asdict(x)  # type: ignore
        dc.astuple(x)  # type: ignore
        dc.replace(x)  # type: ignore

    if dc.is_dataclass(y):
        assert_type(y, Union["DataclassInstance", Type["DataclassInstance"]])
        assert_type(dc.fields(y), Tuple[dc.Field[Any], ...])

        dc.asdict(y)  # type: ignore
        dc.astuple(y)  # type: ignore
        dc.replace(y)  # type: ignore

    if dc.is_dataclass(y) and not isinstance(y, type):
        assert_type(y, "DataclassInstance")
        assert_type(dc.fields(y), Tuple[dc.Field[Any], ...])
        assert_type(dc.asdict(y), Dict[str, Any])
        assert_type(dc.astuple(y), Tuple[Any, ...])
        dc.replace(y)


class _D: ...


custom_dc = dc.dataclass(_D, init=True)
assert_type(custom_dc, type[_D])

custom_dc_2 = dc.dataclass(None, init=True)(_D)
assert_type(custom_dc_2, type[_D])


# Regression test for #11653
D = dc.make_dataclass(
    "D", [("a", Union[int, None]), "y", ("z", Annotated[FrozenSet[bytes], "metadata"], dc.field(default=frozenset({b"foo"})))]
)
# Check that it's inferred by the type checker as a class object of some kind
# (but don't assert the exact type that `D` is inferred as,
# in case a type checker decides to add some special-casing for
# `make_dataclass` in the future)
assert_type(D.__mro__, Tuple[type, ...])


if sys.version_info >= (3, 14):
    from typing import TypeVar

    _T = TypeVar("_T")

    def custom_dataclass(
        cls: type[_T],
        /,
        *,
        init: bool = True,
        repr: bool = True,
        eq: bool = True,
        order: bool = False,
        unsafe_hash: bool = False,
        frozen: bool = False,
        match_args: bool = True,
        kw_only: bool = False,
        slots: bool = False,
        weakref_slot: bool = False,
    ) -> type[_T]:
        custom_dc_maker = dc.dataclass(
            init=init,
            repr=repr,
            eq=eq,
            order=order,
            unsafe_hash=unsafe_hash,
            frozen=frozen,
            match_args=match_args,
            kw_only=kw_only,
            slots=slots,
            weakref_slot=weakref_slot,
        )
        return custom_dc_maker(cls)

    dc.make_dataclass(
        "D",
        [("a", Union[int, None]), "y", ("z", Annotated[FrozenSet[bytes], "metadata"], dc.field(default=frozenset({b"foo"})))],
        decorator=custom_dataclass,
    )
